Imports EnvDTE

Public Class Utils

    <System.Runtime.CompilerServices.MethodImpl(Runtime.CompilerServices.MethodImplOptions.NoInlining)> _
    Public Shared Sub ThrowException(ByVal poExp As Exception)
        Throw poExp
    End Sub

    Public Shared Function GetSingleCodeLine(ByVal poEditPoint As EnvDTE.EditPoint, ByVal plLineNo As Integer) As String
        Return GetSingleCodeLine(poEditPoint, plLineNo, True)
    End Function

    Public Shared Function GetSingleCodeLine(ByVal poEditPoint As EnvDTE.EditPoint, ByVal piLineNo As Integer, ByVal pbTrimLine As Boolean) As String
        Dim sFile As String
        Try
            Dim sReturn As String
            sReturn = poEditPoint.GetLines(piLineNo, piLineNo + 1)
            If pbTrimLine Then
                sReturn = TrimCodeLine(sReturn)
            End If

            Return sReturn
        Catch oExp As Exception
            sFile = poEditPoint.Parent().Parent().ProjectItem().Properties.Item("FullPath").Value.ToString
            ThrowException(New ApplicationException("Error getting line: " + piLineNo.ToString + " in the file: " + sFile, oExp))
        End Try
    End Function

    Public Shared Function TrimCodeLine(ByVal psLine As String) As String
        psLine = psLine.Replace(vbTab, "")
        psLine = psLine.Trim()
        Return psLine
    End Function

    Public Shared Function IsXMLDocLine(ByVal psLine As String) As Boolean
		Dim sCommentString As String
        Dim sFixedLine As String

		Dim oConfig As Configuration = Configuration.GetInstance()
		sCommentString = oConfig.XMLCommentPrefix

        sFixedLine = psLine.Trim

        If sFixedLine.Length >= sCommentString.Length Then
            If sFixedLine.Substring(0, sCommentString.Length) = sCommentString Then 'sCommentString's default value is '''
                Return True
            End If
        End If

        Return False
    End Function

    Public Shared Function IsNonEmptyXMLDocLine(ByVal psLine As String) As Boolean
		Dim sCommentString As String
        Dim sFixedLine As String

		Dim oConfig As Configuration = Configuration.GetInstance()
		sCommentString = oConfig.XMLCommentPrefix

        sFixedLine = psLine.Trim

        If sFixedLine.Length > sCommentString.Length Then
            If sFixedLine.Substring(0, sCommentString.Length) = sCommentString Then 'sCommentString's default value is '''
                Return True
            End If
        End If

        Return False
    End Function

    Public Shared Function IsEmptyXMLDocLine(ByVal psLine As String) As Boolean
		Dim sCommentString As String
        Dim sFixedLine As String

		Dim oConfig As Configuration = Configuration.GetInstance()
		sCommentString = oConfig.XMLCommentPrefix

        sFixedLine = psLine.Trim

        If sFixedLine.Length = sCommentString.Length Then
            If sFixedLine = sCommentString Then 'sCommentString's default value is '''
                Return True
            End If
        End If

        Return False
    End Function

    Public Shared Function GetLastCharsFromString(ByVal psString As String, ByVal piNumberOfChars As Integer) As String
        If psString.Length <= piNumberOfChars Then
            Return psString
        Else
            Return psString.Substring(psString.Length - piNumberOfChars, piNumberOfChars)
        End If
    End Function

    'HACK TextPoint.CodeElement needs quite a lot of help in order to return the correct codeelement
    Public Shared Function GetCodeElementFromCodePosEx(ByVal poEditPoint As EnvDTE.EditPoint, ByRef piMovedLines As Integer) As EnvDTE.CodeElement

        Dim oElement As EnvDTE.CodeElement
        Dim sTmp As String
        Dim bContinue As Boolean = True

        piMovedLines = 0
        oElement = Nothing

        oElement = GetCodeElementFromPointEx(poEditPoint)
        If oElement Is Nothing Then
            ' Nothing found, can we find something by following line continuations ?
            sTmp = GetSingleCodeLine(poEditPoint, poEditPoint.Line, True)
            If Utils.GetLastCharsFromString(sTmp, 2) = " _" Then

                bContinue = True

                While bContinue
                    poEditPoint.EndOfLine()
                    If poEditPoint.AtEndOfDocument Then
                        bContinue = False
                    End If

                    If bContinue Then
                        poEditPoint.LineDown()
                        piMovedLines += 1
                        poEditPoint.StartOfLine()
                        oElement = GetCodeElementFromPointEx(poEditPoint)
                        If oElement Is Nothing Then
                            sTmp = GetSingleCodeLine(poEditPoint, poEditPoint.Line, True)
                            If sTmp.Substring(sTmp.Length - 2, 2) = " _" Then
                                bContinue = True
                            Else
                                bContinue = False
                            End If
                        Else
                            bContinue = False
                        End If
                    End If

                End While
            Else

                sTmp = Utils.GetSingleCodeLine(poEditPoint, poEditPoint.Line, True)
                sTmp = sTmp.Replace(" ", "")
                sTmp = sTmp.Replace(vbTab, "")
                If sTmp.StartsWith("<") Then
                    poEditPoint.LineDown()
                    oElement = GetCodeElementFromPointEx(poEditPoint)
                    poEditPoint.LineUp()
                End If

            End If
        End If

        If Not oElement Is Nothing Then
            ' cache Kind property to avoid COM overhead
            Dim Kind As EnvDTE.vsCMElement = oElement.Kind

            If Kind = vsCMElement.vsCMElementParameter Then
                oElement = CType(oElement, CodeParameter).Parent
            ElseIf Kind = vsCMElement.vsCMElementAttribute Then
                oElement = CType(CType(oElement, CodeAttribute).Parent, EnvDTE.CodeElement)
            End If
        End If

        Return oElement

    End Function

    Public Shared Function GetCodeElementFromPointEx(ByVal oPoint As EnvDTE.TextPoint) As CodeElement

		If Not oPoint.CodeElement(vsCMElement.vsCMElementOther) Is Nothing Then
			'0 An element not in the list. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementOther)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementClass) Is Nothing Then
			'1 A class element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementClass)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementFunction) Is Nothing Then
			'2 A function element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementFunction)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementVariable) Is Nothing Then
			'3 A variable element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementVariable)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementProperty) Is Nothing Then
			'4 A property element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementProperty)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementNamespace) Is Nothing Then
			'5 A namespace element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementNamespace)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementParameter) Is Nothing Then
			'6 A parameter element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementParameter)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementAttribute) Is Nothing Then
			'7 An attribute element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementAttribute)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementInterface) Is Nothing Then
			'8 An interface element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementInterface)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementDelegate) Is Nothing Then
			'9 A delegate element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementDelegate)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementEnum) Is Nothing Then
			'10 An enumerator element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementEnum)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementStruct) Is Nothing Then
			'11 A structure element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementStruct)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementUnion) Is Nothing Then
			'12 A union element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementUnion)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementLocalDeclStmt) Is Nothing Then
			'13 A local declaration statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementLocalDeclStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementFunctionInvokeStmt) Is Nothing Then
			'14 A function invoke statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementFunctionInvokeStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementPropertySetStmt) Is Nothing Then
			'15 A property set statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementPropertySetStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementAssignmentStmt) Is Nothing Then
			'16 An assignment statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementAssignmentStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementInheritsStmt) Is Nothing Then
			'17 An inherits statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementInheritsStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementImplementsStmt) Is Nothing Then
			'18 An implements statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementImplementsStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementOptionStmt) Is Nothing Then
			'19 An option statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementOptionStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementVBAttributeStmt) Is Nothing Then
			'20 A VB attributes statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementVBAttributeStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementVBAttributeGroup) Is Nothing Then
			'21 A VB attribute group element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementVBAttributeGroup)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementEventsDeclaration) Is Nothing Then
			'22 An events declaration element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementEventsDeclaration)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementUDTDecl) Is Nothing Then
			'23 A user-defined type declaration element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementUDTDecl)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementDeclareDecl) Is Nothing Then
			'24 A declare declaration element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementDeclareDecl)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementDefineStmt) Is Nothing Then
			'25 A define statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementDefineStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementTypeDef) Is Nothing Then
			'26 A type definition element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementTypeDef)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementIncludeStmt) Is Nothing Then
			'27 An include statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementIncludeStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementUsingStmt) Is Nothing Then
			'28 A using statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementUsingStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementMacro) Is Nothing Then
			'29 A macro element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementMacro)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementMap) Is Nothing Then
			'30 A map element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementMap)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementIDLImport) Is Nothing Then
			'31 An IDL import element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementIDLImport)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementIDLImportLib) Is Nothing Then
			'32 An IDL import library element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementIDLImportLib)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementIDLCoClass) Is Nothing Then
			'33 An IDL co-class element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementIDLCoClass)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementIDLLibrary) Is Nothing Then
			'34 An IDL library element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementIDLLibrary)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementImportStmt) Is Nothing Then
			'35 An import statement element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementImportStmt)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementMapEntry) Is Nothing Then
			'36 A map entry element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementMapEntry)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementVCBase) Is Nothing Then
			'37 A VC base element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementVCBase)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementEvent) Is Nothing Then
			'38 An event element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementEvent)

		ElseIf Not oPoint.CodeElement(vsCMElement.vsCMElementModule) Is Nothing Then
			'39 A module element. 
			Return oPoint.CodeElement(vsCMElement.vsCMElementModule)
        Else

            ' Commented out by Tor-Erik 28.02.2004, this code does more harm than good
            '
            '''''''''         ' worse case scenario - handle problem when we have text after the declaration
            '''''''''         ' on the same line.
            '''''''''         ' 
            '''''''''         ' eg.
            '''''''''         '  Public Class MyException ' My comment
            '''''''''         '  Public Class MyException : Inherits Exception
            '''''''''         '  Public Class MyException : Implements MyInterface

            '''''''''         Dim objCodeElement As CodeElement

            '''''''''         ' TODO: Should we recurse deeper than just the top-level CodeElements collection?
            '''''''''         For Each objCodeElement In oPoint.Parent.Parent.ProjectItem.ContainingProject.CodeModel.CodeElements

            '''''''''             ' .GetStartPoint() is not implemented for some reason
            '''''''''             Dim sp As TextPoint = objCodeElement.StartPoint '.GetStartPoint() '(vsCMPart.vsCMPartName)
            '''''''''             Dim np As TextPoint = objCodeElement.EndPoint '.GetEndPoint() 'vsCMPart.vsCMPartName)

            '''''''''             ' The np.LineCharOffset doesn't make sense (bug?), so we do without it.
            '''''''''	If ((sp.Line <= oPoint.Line) AndAlso (np.Line >= oPoint.Line) AndAlso (sp.LineCharOffset <= oPoint.LineCharOffset)) Then

            '''''''''		Return (objCodeElement)
            '''''''''	End If

            'Next
            Return Nothing
        End If

    End Function

    Public Shared Function ApplyXSLTransformation(ByVal psXMLDocument As String, ByVal psXSLTDocument As String) As String

        Dim oXSLTStream As IO.MemoryStream
        Dim oXMLStream As IO.MemoryStream
        Dim oTransform As Xml.Xsl.XslTransform
        Dim oDocument As Xml.XPath.XPathDocument
        Dim oWriter As IO.StringWriter

        oXSLTStream = New IO.MemoryStream(System.Text.Encoding.UTF8.GetBytes(psXSLTDocument))
        oXMLStream = New IO.MemoryStream(System.Text.Encoding.UTF8.GetBytes(psXMLDocument))


        oTransform = New Xml.Xsl.XslTransform()

#If VBNET2002 Then
        oTransform.Load(New Xml.XmlTextReader(oXSLTStream))
#Else
        oTransform.Load(New Xml.XmlTextReader(oXSLTStream), Nothing, Nothing)

#End If

        oDocument = New Xml.XPath.XPathDocument(oXMLStream)

        oWriter = New IO.StringWriter

#If VBNET2002 Then
        oTransform.Transform(oDocument, Nothing, oWriter)
#Else
        oTransform.Transform(oDocument, Nothing, oWriter, Nothing)

#End If

		Return oWriter.ToString.Trim()

    End Function

    Public Shared Sub XMLCommentifyLines(ByVal psCommentStub As System.Text.StringBuilder)

		Dim oConfig As Configuration = Configuration.GetInstance()
		Dim PaddedCommentString As String = vbCrLf & oConfig.XMLCommentPrefix & " "

        ' insert leading space
        psCommentStub.Insert(0, " ")

        ' replace line-breaks with line-break + xml comment prefix
        psCommentStub.Replace(vbCrLf, PaddedCommentString)

    End Sub

    Public Shared Function GetResourceStream(ByVal psResourceName As String) As IO.Stream

        Dim oAssembly As Reflection.Assembly
        Dim oStream As IO.Stream

        oAssembly = Reflection.Assembly.GetExecutingAssembly()
        oStream = oAssembly.GetManifestResourceStream(psResourceName)

        Return oStream

    End Function

    Public Shared Function ParseCodeAttribute(ByVal psAttributeValue As String) As Hashtable

        Dim bInsideString As Boolean
        Dim bPossibleEndOfString As Boolean
        Dim iPos As Integer
        Dim sBuffer As String
        Dim oRes As Hashtable
        Dim sCurName As String
        Dim sCurValue As String

        oRes = New Hashtable()
        sBuffer = ""
        sCurName = ""
        sCurValue = ""

        For iPos = 0 To psAttributeValue.Length - 1

            Select Case psAttributeValue.Substring(iPos, 1)
                Case Chr(34)
                    If Not bInsideString Then
                        bInsideString = True
                        If sBuffer.Trim.Length > 0 Then
                            sCurName = sBuffer.Trim
                        End If
                        sBuffer = ""
                    Else
                        If bPossibleEndOfString Then
                            bPossibleEndOfString = False
                            sBuffer += Chr(34)
                        Else
                            bPossibleEndOfString = True
                        End If
                    End If
                Case Else
                    If bPossibleEndOfString Then
                        ' We have end of string
                        sCurValue = sBuffer
                        If sCurName.Length > 0 Then
                            oRes.Add(sCurName.ToLower, sCurValue)
                        Else
                            oRes.Add(oRes.Count + 1, sCurValue)
                        End If
                        sCurName = ""
                        sCurValue = ""
                        sBuffer = ""

                        If Not (psAttributeValue.Substring(iPos, 1) = ":" Or _
                               psAttributeValue.Substring(iPos, 1) = "," Or _
                               psAttributeValue.Substring(iPos, 1) = " " Or _
                               psAttributeValue.Substring(iPos, 1) = "=") Then
                            sBuffer += psAttributeValue.Substring(iPos, 1)
                        End If

                        bPossibleEndOfString = False
                        bInsideString = False
                    Else
                        If bInsideString Then
                            sBuffer += psAttributeValue.Substring(iPos, 1)
                        Else
                            If Not (psAttributeValue.Substring(iPos, 1) = ":" Or _
                                   psAttributeValue.Substring(iPos, 1) = "," Or _
                                   psAttributeValue.Substring(iPos, 1) = " " Or _
                                   psAttributeValue.Substring(iPos, 1) = "=") Then
                                sBuffer += psAttributeValue.Substring(iPos, 1)
                            End If
                        End If
                    End If
            End Select

        Next

        If bPossibleEndOfString Then
            ' We have end of string
            sCurValue = sBuffer
            If sCurName.Length > 0 Then
                oRes.Add(sCurName.ToLower, sCurValue)
            Else
                oRes.Add(oRes.Count + 1, sCurValue)
            End If
        End If

        Return oRes

    End Function

    'HACK It's not easy to get the parameters from a code element
    Public Shared Function GetCodeElementParameters(ByVal poCodeElement As CodeElement) As EnvDTE.CodeParameter()
        Dim oRet As EnvDTE.CodeParameter()
        Dim oFunc As EnvDTE.CodeFunction
        Dim oProp As EnvDTE.CodeProperty
        Dim oPara As EnvDTE.CodeParameter
        Dim oRetList As Collection
        Dim iPos As Integer
        Dim bIgnoreParamsWithoutStartPoint As Boolean

        If poCodeElement.Kind = vsCMElement.vsCMElementFunction Or _
           poCodeElement.Kind = vsCMElement.vsCMElementEventsDeclaration Or _
           poCodeElement.Kind = vsCMElement.vsCMElementProperty Then

            oRetList = New Collection
            bIgnoreParamsWithoutStartPoint = False

            If poCodeElement.Kind = vsCMElement.vsCMElementFunction Or _
               poCodeElement.Kind = vsCMElement.vsCMElementEventsDeclaration Then
                ' Function
                oFunc = CType(poCodeElement, CodeFunction)
            Else
                ' Property
                oProp = CType(poCodeElement, CodeProperty)

                If Not oProp.Getter Is Nothing Then
                    oFunc = oProp.Getter
                Else
                    bIgnoreParamsWithoutStartPoint = True
                    oFunc = oProp.Setter
                End If
            End If

            For Each oPara In oFunc.Parameters
                If Not ((oPara.StartPoint Is Nothing) AndAlso bIgnoreParamsWithoutStartPoint) Then
                    If bIgnoreParamsWithoutStartPoint Then
                        ' We know that the parameters is defined outside the setter, so we ignore any parameter that is set inside the setter
                        If Not oFunc.StartPoint Is Nothing Then
                            If Not oPara.StartPoint.Line >= oFunc.StartPoint.Line Then
                                oRetList.Add(oPara)
                            End If
                        Else
                            oRetList.Add(oPara)
                        End If
                    Else
                        oRetList.Add(oPara)
                    End If

                End If
            Next

            ReDim oRet(oRetList.Count - 1)
            For iPos = 0 To oRetList.Count - 1
                oRet(iPos) = CType(oRetList.Item(iPos + 1), EnvDTE.CodeParameter)
            Next

            Return oRet

        Else
            Return Nothing
        End If

    End Function

    'HACK To tell if a parameter is byref or not is not simple
    Public Shared Function IsVBByRefParameter(ByVal poCodeParameter As EnvDTE.CodeParameter) As Boolean
        Dim sText As String
        Dim oEdit As EditPoint

        oEdit = poCodeParameter.StartPoint().CreateEditPoint()
        sText = oEdit.GetText(poCodeParameter.EndPoint())
        sText = sText.Trim()

        If sText.ToLower.StartsWith("byref ") Then
            Return True
        Else
            If sText.ToLower.IndexOf(" byref ") <> -1 Then
                Return True
            Else
                Return False
            End If
        End If

    End Function

    Public Shared Function GetParentsName(ByVal poCodeElement As EnvDTE.CodeElement) As String

        Dim sName As String
        Dim iDotPos As Integer
        sName = poCodeElement.FullName

        iDotPos = sName.LastIndexOf(".")
        If iDotPos > 0 Then
            Return sName.Substring(0, iDotPos)
        Else
            Return ""
        End If

    End Function

    ' HACK this is really naaasty
    Public Shared Function IsEnumerationMember(ByVal poCodeElement As EnvDTE.CodeElement) As Boolean

        ' There is no way to tell the difference between a variable and a enum member,
        ' so we have to check if it's parent is a enumeration. But the parent property
        ' is always nothing. Therefore we have to get it's parent by it's name

        Dim sParentName As String
        Dim oCodeElement As EnvDTE.CodeElement

        If poCodeElement.Kind = vsCMElement.vsCMElementVariable Then

            sParentName = GetParentsName(poCodeElement)
            If sParentName.Length > 0 Then

                Dim ct As CodeType = poCodeElement.ProjectItem.ContainingProject.CodeModel.CodeTypeFromFullName(sParentName)

                If ct.Kind = vsCMElement.vsCMElementEnum Then
                    Return True
                Else
                    Return False
                End If
            Else
                Return False
            End If
        Else
            Return False
        End If

    End Function

    Public Shared Sub CreateXPathInXMLDocument(ByVal psPath As String, ByVal poDocument As Xml.XmlDocument)

        Dim oActiveNode As Xml.XmlNode
        Dim aPath() As String
        Dim iNode As Integer
        Dim oNodeSearch As Xml.XmlNode

        ' Verify that the string has
        If Not psPath = "" Then
            psPath = psPath.Trim()
            If psPath = "" Then
                Exit Sub
            End If
        Else
            Exit Sub
        End If

        oActiveNode = poDocument.DocumentElement
        aPath = Split(psPath, "/")

        aPath = removeEmptyElementsFromArray(aPath)

        If Not aPath(0) = oActiveNode.Name Then
            Throw New ApplicationException("There already exist a root element with a different name than the one specified in the xpath")
        End If

        For iNode = 1 To aPath.Length - 1
            oNodeSearch = oActiveNode.SelectSingleNode("./" + aPath(iNode))
            If oNodeSearch Is Nothing Then
                ' TODO test if there does exist multiple nodes!
                oActiveNode = oActiveNode.AppendChild(oActiveNode.OwnerDocument.CreateNode(Xml.XmlNodeType.Element, aPath(iNode), ""))
            Else
                oActiveNode = oNodeSearch
            End If
        Next

    End Sub

    Public Shared Function removeEmptyElementsFromArray(ByVal paArray() As String) As String()

        Dim iMainCnt As Integer
        Dim iRetCnt As Integer
        Dim aRet() As String

        ReDim aRet(paArray.Length - 1)
        iRetCnt = 0

        For iMainCnt = 0 To paArray.Length - 1
            If paArray(iMainCnt) <> "" Then
                aRet(iRetCnt) = paArray(iMainCnt)
                iRetCnt += 1
            End If
        Next

        If iRetCnt > 0 Then
            ReDim Preserve aRet(iRetCnt - 1)
        Else
            aRet = Nothing
        End If

        Return aRet

    End Function

    ''' <summary>
    ''' String.TrimStart() for StringBuilder objects
    ''' </summary>
    Public Shared Sub TrimStart(ByVal text As System.Text.StringBuilder, ByVal ParamArray trimChars As Char())

        If (text.Length > 0) Then
            Dim i As Integer = 0
            Dim length As Integer = text.Length

            ' make string out of chars
            Dim trimString As String

            If trimChars.Length <= 0 Then
                trimString = vbTab & vbCr & vbLf & " "
            Else
                Dim ch As Char
                For Each ch In trimChars
                    trimString &= ch
                Next

            End If

            ' find out how far along we need to prune
            Do While (i < text.Length) AndAlso (trimString.IndexOf(text.Chars(i)) >= 0)
                i += 1
            Loop

            If (i > 0) Then
                text.Remove(0, i)
            End If

            'length = text.Length
            'i = length - 1
            'Do While (i >= 0) AndAlso ((text.Chars(i) = BORDER_CHAR) OrElse (text.Chars(i) = " ") OrElse (text.Chars(i) = vbCr) OrElse (text.Chars(i) = vbLf) OrElse (text.Chars(i) = vbTab))
            '    i -= 1
            'Loop

            'If ((length - 1 - i) > 0) Then
            '    text.Remove(i, length - 1 - i)
            'End If

        End If
    End Sub

    Public Shared Sub SplitSeparatedValues(ByVal psString As String, ByVal pcSeparator As Char, ByRef psPart1 As String, ByRef psPart2 As String)

        Dim iPos As Integer

        iPos = psString.IndexOf(pcSeparator)

        psPart1 = ""
        psPart2 = ""

        If iPos = -1 Then
            psPart1 = psString
        Else
            psPart1 = psString.Substring(0, iPos - 1)
            If iPos < psString.Length - 1 Then
                psPart2 = psString.Substring(iPos + 1)
            End If
        End If

    End Sub

    Public Shared Function StringStackToSeparatedString(ByVal poStringStack As Stack, Optional ByVal psSeparator As String = "") As String

        Dim sCurString As String = ""
        Dim sTotalString As String = ""

        For Each sCurString In poStringStack
            sTotalString = sCurString & psSeparator & sTotalString
        Next

        Return sTotalString

    End Function

End Class
