Option Strict On
Option Explicit On 

Imports EnvDTE
Imports System.Windows.Forms
Imports System.Text.RegularExpressions

Public Class VBCommenter

    Public Const SIGNATURE_MODIFIERS As String = "(?mi)((<.*>|Dim|Const|Overloads|Default|Overrides|Overridable|NotOverridable|MustOverride|Shadows|(?<shared>Shared)|Public|Protected|Friend|Protected\s+Friend|Private|ReadOnly|(?<withEvents>WithEvents)|Serializable|WriteOnly|MustInherit|NotInheritable)\s+)*((?<delegate>Delegate)\s+)?(?<type>\w*)\s*\[?(?<name>[_|a-z]\w*)\]?\s*(?<params>.*)"

	Public Sub New()
	End Sub

    Private Function validateTextPoint(ByVal point As EnvDTE.TextPoint) As Boolean
        Try
            Dim test As Object = point.Parent.Parent.ActiveWindow.Selection
            test = point.Parent.Parent.FullName
        Catch ex As Exception
            Return False
        End Try
        Return True
    End Function

    Public Sub Commentate(ByVal poStartPoint As EnvDTE.TextPoint, ByVal poEndPoint As EnvDTE.TextPoint, ByVal piHint As Integer)

        Dim oEditPoint As EditPoint
        Dim sTmp As String
        Dim oTextInsert As TextInsertion
        Dim iEndLine As Integer
        Dim bFinishedHandling As Boolean = False
        Dim sCursorLine As String
        If Me.validateTextPoint(poStartPoint) = True And _
            Me.validateTextPoint(poEndPoint) = True Then
            Try

                ' Check that this is a VB file
                If Not bFinishedHandling Then
                    If Not poStartPoint.Parent.Parent.FullName.ToLower.EndsWith(".vb") Then
                        bFinishedHandling = True
                    End If
                End If

                'poStartPoint.Parent.Parent.ActiveWindow
                'Dim ActivePoint As VirtualPoint = CType(poStartPoint.DTE.ActiveDocument.Selection, TextSelection).ActivePoint
                Dim ActivePoint As VirtualPoint = CType(poStartPoint.Parent.Parent.ActiveWindow.Selection, TextSelection).ActivePoint

                ' Check that we have just had a newline and that the previous line is an XML DOC comment
                If Not bFinishedHandling Then
                    If (piHint And EnvDTE.vsTextChanged.vsTextChangedNewline) = EnvDTE.vsTextChanged.vsTextChangedNewline Then
                        oEditPoint = ActivePoint.CreateEditPoint()    ' poEndPoint.CreateEditPoint()
                        If poEndPoint.Line = poStartPoint.Line Then
                            ' User hasn't pressed enter, and we don't want to handle just ''' and then moving out of the line
                            bFinishedHandling = True
                        Else
                            iEndLine = oEditPoint.Line       ' poEndPoint.Line
                        End If

                        If Not bFinishedHandling Then
                            sTmp = Utils.GetSingleCodeLine(oEditPoint, iEndLine - 1, True)
                            sCursorLine = Utils.GetSingleCodeLine(oEditPoint, iEndLine, True)

                            If Not Utils.IsXMLDocLine(sTmp) Then
                                bFinishedHandling = True
                            End If
                        End If
                    Else

                        bFinishedHandling = True
                    End If

                End If

                ' Check if the user has pressed enter inside an XML Doc block, in that case add ''' to the start of the new line
                If Not bFinishedHandling Then
                    bFinishedHandling = CheckIfNewlineInsideComment(oEditPoint, iEndLine, oTextInsert)
                End If

                ' Check the next line if it is the first line of
                ' something that support's XML Doc comments
                '
                ' The code for this is messy, but the objects we work against
                ' aren't precise ...
                If Not bFinishedHandling Then
                    bFinishedHandling = CheckIfStubShouldBeInserted(oEditPoint, oEditPoint, iEndLine, sCursorLine, oTextInsert)
                End If

                ' Insert any code that should be inserted
                If Not oTextInsert Is Nothing Then
                    If oTextInsert.HasAnythingToAdd() Then
                        oTextInsert.PerformInsertion()
                    End If
                End If

                ' Validate XML and add any errors to task list.
                'ValidateXML()

            Catch oExp As Exception
                'MsgBox(oExp.ToString)
            End Try
        End If        
    End Sub



    Private Function CheckIfStubShouldBeInserted(ByVal poEditPoint As EnvDTE.EditPoint, ByVal poEndPoint As EnvDTE.EditPoint, ByVal piEndLine As Integer, ByVal psCursorLine As String, ByRef poTextInsert As TextInsertion) As Boolean
        Dim sTmp As String
        Dim iMovedLines As Integer
        Dim oTmpEditPoint As EnvDTE.EditPoint
        Dim oTmpPoint As EnvDTE.EditPoint
        Dim oEle As EnvDTE.CodeElement
        Dim bCheckCodePos As Boolean
        Dim bDone As Boolean
        Dim bRet As Boolean

        bRet = False

        sTmp = Utils.GetSingleCodeLine(poEditPoint, piEndLine - 1, True)

        If Utils.IsEmptyXMLDocLine(sTmp) Then

            iMovedLines = 0
            oTmpEditPoint = poEndPoint.CreateEditPoint()
            oTmpEditPoint.LineDown()
            oTmpEditPoint.StartOfLine()
            oEle = Utils.GetCodeElementFromCodePosEx(oTmpEditPoint, iMovedLines)

            ' Check that we're not inside something
            If Not oEle Is Nothing Then

                bCheckCodePos = True

                'TODO: Implement full parsing of the enumeration to verify the linenumber
                ' The StartPoint and EndPoint of a Enum variable is reported wrong, so we skip the "correct line" test
                ' But as an enumeration member fails to return it's parent we have to do some tricks in order to
                ' verify that it is an enumeration member and not an ordinary member
                If Utils.IsEnumerationMember(oEle) Then
                    bCheckCodePos = False
                End If

                If bCheckCodePos Then
                    ' oTmpPoint = oEle.GetStartPoint(vsCMPart.vsCMPartHeader)
                    ' This works better, the line above returns the last non-available line and get's messed up with any attribute decorations, inheritance etc ...
                    oTmpPoint = oEle.StartPoint.CreateEditPoint()

                    ' Cache .Line property to avoid COM overhead
                    Dim TmpPointLine As Integer = oTmpPoint.Line

                    ' HACK (unclean due to bug) Line number is incorrect when dealing with classes, functions works OK
                    If iMovedLines = 0 Then
                        If Not ((TmpPointLine = piEndLine) OrElse (TmpPointLine = piEndLine + 1)) Then
                            bDone = True
                        End If
                    Else
                        ' We have moved the edit point in order to follow line continuation (possibly because of attributes)
                        ' but we might get it returned with or without the continuation .. and with or without the added line ...
                        If Not (TmpPointLine = (piEndLine - iMovedLines)) AndAlso _
                         Not (TmpPointLine = ((piEndLine + 1) - iMovedLines)) AndAlso _
                          Not (TmpPointLine = ((piEndLine + 1) + iMovedLines)) AndAlso _
                          Not (TmpPointLine = piEndLine) AndAlso _
                          Not (TmpPointLine = (piEndLine + 1)) Then       ' This should be the correct case ... but we'll leave the others too for now
                            ' No luck .. we have got something wrong in return ...
                            bDone = True
                        End If
                    End If
                End If
            Else
                bDone = True
            End If

            If Not bDone Then
                ' Generate XML comment stub

                poTextInsert = New TextInsertion(poEditPoint)

                poTextInsert.MoveLinesDownPreInsert = -1
                If Not psCursorLine = "" Then
                    ' NB! It's important to reset this after setting MoveLinesDownPreInsert as MoveLinesDownPreInsert
                    '     internally sets SoftNewlineCount if MoveLinesDownPreInsert has a negative value
                    poTextInsert.SoftNewlineCount = 0
                End If
                poTextInsert.ColumnMoveModePreInsert = ColumnMoveMode.icMoveToEndOfLine
                poTextInsert.ColumnMoveModeAfterInsert = ColumnMoveMode.icMoveToEndOfLine

                GenXMLDocStub(poTextInsert, oEle)

                bRet = True
            End If

        End If

        If (Not bDone) OrElse bRet Then
            Return True
        Else
            Return False
        End If


    End Function

    Private Function CheckIfNewlineInsideComment(ByVal poEditPoint As EnvDTE.EditPoint, ByVal piEndLine As Integer, ByRef poTextInsert As TextInsertion) As Boolean
        Dim bInsertXMLCommentStart As Boolean
        Dim sTmp As String
        Dim sCommentString As String
        Dim bRet As Boolean

        Dim oConfig As Configuration = Configuration.GetInstance()
        sCommentString = oConfig.XMLCommentPrefix

        bInsertXMLCommentStart = False
        bRet = False

        If Not poEditPoint.Line <= 2 Then
            sTmp = Utils.GetSingleCodeLine(poEditPoint, piEndLine - 2, True)
            If Utils.IsXMLDocLine(sTmp.Trim) Then
                bInsertXMLCommentStart = True
            Else
                sTmp = Utils.GetSingleCodeLine(poEditPoint, piEndLine - 1, True)
                If Not Utils.IsEmptyXMLDocLine(sTmp.Trim) Then
                    bInsertXMLCommentStart = True
                End If
            End If
        Else
            sTmp = Utils.GetSingleCodeLine(poEditPoint, piEndLine - 1, True)
            If Not Utils.IsEmptyXMLDocLine(sTmp.Trim) Then
                bInsertXMLCommentStart = True
            End If
        End If

        If bInsertXMLCommentStart Then
            poTextInsert = New TextInsertion(poEditPoint)
            poTextInsert.ColumnMoveModePreInsert = ColumnMoveMode.icStartAtTextStart
            poTextInsert.ColumnMoveModeAfterInsert = ColumnMoveMode.icStartAtTextStart
            poTextInsert.MoveColumnsAfterInsert = sCommentString.Length + 1
            poTextInsert.AddLine(sCommentString + " ")    'sCommentString's default value is '''
            ' We have just added the needed text so we stop handling the newline
            bRet = True
        End If

        Return bRet

    End Function


    Private Sub GenXMLDocStub(ByVal poTextInsert As TextInsertion, ByVal poCodeElement As EnvDTE.CodeElement)

        Dim iLine As Integer
        Dim iLineCount As Integer
        Dim iCol As Integer
        Dim aLines As String()
        Dim bFound As Boolean
        Dim iCurLine As Integer

        Dim sStub As New System.Text.StringBuilder(Template.GetFilledXMLTemplateForCodeElement(TemplateTypesEnum.icDocStubTemplate, poCodeElement))

        Utils.TrimStart(sStub)

        sStub.Replace("[CRLF]", vbCrLf)
        sStub.Replace("[TAB]", vbTab)
        sStub.Replace("[TRICK]", "")

        ' Prefix each line with XML Comment prefix
        Utils.XMLCommentifyLines(sStub)

        FindCursorPos(sStub.ToString(), iLine, iCol)

        bFound = False

        sStub.Replace("[CURSOR]", "")

        aLines = sStub.Replace(vbCrLf, vbCr).ToString().Split(Chr(13))

        iLineCount = aLines.Length

        For iCurLine = 0 To aLines.Length - 1
            poTextInsert.AddLine(aLines(iCurLine))
        Next

        poTextInsert.ColumnMoveModeAfterInsert = ColumnMoveMode.icStartAtTextStart
        poTextInsert.MoveColumnsAfterInsert = iCol - 1

        poTextInsert.MoveLinesDownAfterInsert = (iLine - iLineCount) + 1

    End Sub

    Private Sub FindCursorPos(ByVal psCommentStub As String, ByRef piLine As Integer, ByRef piColumn As Integer)

        Dim aLines As String()
        Dim bFound As Boolean
        Dim iCurLine As Integer

        bFound = False
        aLines = psCommentStub.Replace(vbCrLf, vbCr).Split(Chr(13))

        For iCurLine = 0 To aLines.Length - 1
            If aLines(iCurLine).IndexOf("[CURSOR]") <> -1 Then
                piLine = iCurLine
                piColumn = aLines(iCurLine).IndexOf("[CURSOR]") + 1
                bFound = True
                Exit For
            End If
        Next

        If Not bFound Then
            piColumn = aLines(aLines.Length - 1).Length
            piLine = aLines.Length
        End If

    End Sub

End Class