Option Strict On
Option Explicit On 

Imports EnvDTE
Imports System.IO
Imports System.Xml
Imports System.Reflection
Imports System.Windows.Forms
'Imports System.Reflection.Emit
Imports System.Text.RegularExpressions
Imports System.Text

''' <summary>
''' Summary description for XMLWriter.
''' </summary>
Public Class XmlWriter

#Region " Member Variables "

    Protected Const m_PARAMS_EXPR As String = "(?imn)[(,]\s*(Optional|ByVal|ParamArray|\s*)*(?<byref>ByRef|\s*)\s*(?<name>[a-z]\w*)(?<array1>\(.*?\)|\s*)\s*(as\s(?<type>[a-z][\w\.]*)\s*(?<array2>\(.*?\)|\s*)|\s*)"
    Protected Const m_HEADER_EXPR As String = "(?i)\<(?<tag>[^\>]+)\>(?<content>.*)\</[^\>]\>"
    Protected Const m_SUMMARY_EXPR As String = "(?i)<summary>(?<summary>.*)</summary>"
    Protected m_CREF_EXPR As String = "(?in)(cref\s*=\s*""(?<cref>[^""]*)"")"
    ' Protected Const m_CREF_EXPR2 As String = "cref\s*=\s*(?:""(?<cref>[^""]*)""|(?<1>\S+))"
    Protected Const m_USER_SUPPLIED_MARKER_EXPR As String = "^[TMEFP]:"
    Protected Const m_CONSTRUCTOR As String = ".new"
    Protected Const XML_CONSTRUCTOR As String = ".#ctor"
    Protected Const XML_STATIC_CONSTRUCTOR As String = ".#cctor"
    Protected Const TaskListCategory As String = "XML Comment"

    Protected m_projectNamespace As String
    Protected m_projectPath As String
    Protected m_projectOutputPath As String
    Protected m_intermediatePath As String
    Protected m_assemblyName As String
    Protected m_confStr As String
    Protected m_platformStr As String

    'Could utilize these for Class matching logic
    'protected Regex m_classBegin = new Regex("(?mi)class\s+(?<name>\S+)");
    'protected Regex m_classEnd = new Regex("(?mi)$\s*end\s+class");
    'protected Match m_beginMatch, m_endMatch;	

    Protected m_xmlWriter As XmlTextWriter
    Protected m_logWriter As StreamWriter

    Protected m_project As Project
    Protected m_assembly As [Assembly]
    Protected m_topLevelTypes() As Type
    Protected m_allTypes As ArrayList
    Protected m_curFileProcessing As String

    Protected m_appDomain As AppDomain

    Protected m_namespaces As New Stack
    Protected m_classes As New Stack
    Protected m_enums As New Stack

    Protected m_localNamespaces As NamespaceCollection
    Protected m_importedNamespaces As NamespaceCollection
    Protected m_VBPrimitivesToConvert As New Hashtable

    Protected m_appObj As EnvDTE.DTE

    Protected myschema As New System.Xml.Schema.XmlSchemaCollection
    Protected mLineNumber As Integer
    Protected mCommentStartLineNumber As Integer
    Protected tasklist As EnvDTE.TaskList
    Protected haveTaskList As Boolean

#End Region

#Region " Constructor "

    ''' <summary>
    ''' Construct an XMLWriter object
    ''' </summary>
    ''' <param name="appObj">DTE app object</param>
    Public Sub New(ByVal appObj As EnvDTE.DTE)
        Try
            m_appObj = appObj
            haveTaskList = True

            Try
                tasklist = DirectCast(m_appObj.Windows.Item(Constants.vsWindowKindTaskList).Object, EnvDTE.TaskList)
            Catch ex As Exception
                haveTaskList = False
            End Try

            ' Not "System." to save time
            m_VBPrimitivesToConvert.Add("short", "Int16")
            m_VBPrimitivesToConvert.Add("integer", "Int32")
            m_VBPrimitivesToConvert.Add("long", "Int64")
            m_VBPrimitivesToConvert.Add("date", "DateTime")

            'Add the schema for validating XML
            Dim oConfig As Configuration = Configuration.GetInstance()

            Dim reader As New System.Xml.XmlTextReader(Utils.GetResourceStream("VBCommenter.comment.xsd"))
            myschema.Add("", reader)
            reader.Close()

        Catch exp As Exception
            MessageBox.Show(exp.Message, "VBCommenter: Error setting DTE object", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End Try
    End Sub

#End Region

#Region " ParseFile Method "

    ''' <summary>
    ''' This method is called on each file in the Project that contains comments.
    ''' </summary>
    ''' <param name="reader">StreamReader for the file to parse comments from</param>

    Private Function SplitMultiLineStatement(ByVal psLine As String) As System.Collections.Specialized.StringCollection

        Dim iPos As Integer
        Dim scReplacer As String = "__VBCOMMENTER__" + "COLON_EQUAL__"

        'Remove any double quotes
        psLine = psLine.Replace("""""", """")

        'Remove the content of any strings
        psLine = System.Text.RegularExpressions.Regex.Replace(psLine, """.*?""", "")

        'Remove any comments
        psLine = System.Text.RegularExpressions.Regex.Replace(psLine, "'.*", "")

        Dim oRes As New System.Collections.Specialized.StringCollection

        ' := is a special case we have to get out of the way
        psLine = psLine.Replace(":=", scReplacer)

        oRes.AddRange(psLine.Split(":"c))

        For iPos = 0 To oRes.Count - 1
            If oRes.Item(iPos).IndexOf(scReplacer) >= 0 Then
                oRes.Item(iPos) = oRes.Item(iPos).Replace(scReplacer, ":=")
            End If
        Next

        Return oRes

    End Function


    Protected Sub ParseFile(ByVal reader As StreamReader)

        Const BORDER_CHAR As String = "-"

#If DEBUG Then
        Const ANYGLOBALO As String = "anyglobalo"
#End If


        ' Remove old tasks for this file
        If haveTaskList = True Then
            Dim task As EnvDTE.TaskItem

            For Each task In tasklist.TaskItems
                If task.Category = TaskListCategory AndAlso task.FileName = Me.m_curFileProcessing Then
                    task.Delete()
                End If
            Next
        End If


        Try
            Dim tempCurLine As String
            Dim signature As String = ""
            Dim curLine As String

            Dim oConfig As Configuration = Configuration.GetInstance()

            Dim sCommentString As String = oConfig.XMLCommentPrefix

            Dim header As New System.Text.StringBuilder

            Dim SubLines As New System.Collections.Specialized.StringCollection

            curLine = reader.ReadLine()
            mLineNumber = 1

            While Not curLine Is Nothing
                curLine = curLine.Trim()

                signature = ""

#If DEBUG Then
                If curLine.IndexOf("anyglobalo") >= 0 Then
                    Debug.WriteLine(curLine)
                    Console.WriteLine("")
                End If
#End If

                If curLine.StartsWith(sCommentString) And curLine.IndexOf("<") >= 0 Then
                    'Header Found - collect header

                    mCommentStartLineNumber = mLineNumber

                    header.Length = 0

                    While (Not curLine Is Nothing) AndAlso (curLine.Trim.StartsWith(sCommentString) OrElse curLine.Trim().StartsWith("'") OrElse (curLine.Trim() = ""))

#If DEBUG Then
                        If curLine.IndexOf(ANYGLOBALO) >= 0 Then
                            Debug.WriteLine(curLine)
                            Console.WriteLine("")
                        End If
#End If

                        If curLine.Trim().StartsWith(sCommentString) Then
                            'header.Append(curLine.Trim.TrimStart((vbTab & vbCrLf & "' ").ToCharArray))
                            curLine = curLine.TrimStart(" "c)
                            If curLine.Length > sCommentString.Length Then
                                curLine = curLine.Substring(sCommentString.Length)
                            Else
                                curLine = ""
                            End If
                            header.Append(curLine)
                            If Not header.Chars(header.Length - 1) = ">"c Then
                                header.Append(" ")
                            End If
                            header.Append(vbCrLf)
                        End If
                        curLine = reader.ReadLine().Trim()
                        mLineNumber += 1
                    End While

                    ' Replace <notes></notes> tags with remarks for backwards compatibilty						
                    header.Replace("<notes>", "<remarks>")
                    header.Replace("</notes>", "</remarks>")

                    ' Trim special characters from header
                    If (header.Length > 0) Then
                        Dim i As Integer = 0
                        Dim length As Integer = header.Length

                        Do While (i < header.Length) AndAlso ((header.Chars(i) = BORDER_CHAR) OrElse (header.Chars(i) = " ") OrElse (header.Chars(i) = vbCr) OrElse (header.Chars(i) = vbLf) OrElse (header.Chars(i) = vbTab))
                            i += 1
                        Loop

                        If (i > 0) Then
                            header.Remove(0, i)
                        End If

                        length = header.Length
                        i = length - 1
                        Do While (i >= 0) AndAlso ((header.Chars(i) = BORDER_CHAR) OrElse (header.Chars(i) = " ") OrElse (header.Chars(i) = vbCr) OrElse (header.Chars(i) = vbLf) OrElse (header.Chars(i) = vbTab))
                            i -= 1
                        Loop

                        If ((length - 1 - i) > 0) Then
                            header.Remove(i + 1, length - 1 - i)
                        End If

                    End If

                    If Not header.Chars(0) = ("<") Then
                        ' Trim to first "<"

                        Dim i As Integer = 0
                        Dim length As Integer = header.Length

                        Do While (i < length) AndAlso (header.Chars(i) <> "<")
                            i += 1
                        Loop

                        header.Remove(0, i)

                    End If

                    ' Handle colon separated statements by creating
                    ' a collection of pseudo-lines using the colon as a separator
                    SubLines.Clear()
                    SubLines = SplitMultiLineStatement(curLine)


                    ' grab the first (or only) line
                    curLine = SubLines.Item(0)
                    SubLines.RemoveAt(0)

                    While (Not curLine Is Nothing) AndAlso ((curLine.StartsWith("'") OrElse curLine = ""))
#If DEBUG Then
                        If curLine.IndexOf(ANYGLOBALO) >= 0 Then
                            Debug.WriteLine(curLine)
                            Console.WriteLine("")
                        End If
#End If
                        curLine = reader.ReadLine()       '.Trim()
                        mLineNumber += 1
                    End While

                    'Collect signature
                    signature = curLine.Trim()
                    While Not signature Is Nothing AndAlso signature.TrimEnd.EndsWith("_")
#If DEBUG Then
                        If signature.IndexOf(ANYGLOBALO) >= 0 Then
                            Debug.WriteLine(curLine)
                            Console.WriteLine("")
                        End If
#End If
                        signature = signature.Trim((" " & vbCrLf + vbTab + "_").ToCharArray()) & " " & reader.ReadLine()
                        mLineNumber += 1
                        'Fix for problem of class attributes and line continuation
                        curLine = signature
                    End While

                    WriteXML(header, signature)
                End If

                Const NAMESPACE_STMT As String = "namespace"
                Const END_NAMESPACE_STMT As String = "endnamespace"
                Const END_STMT As String = "end"

                Dim delimStr As String = " "
                Dim delimiter As Char() = delimStr.ToCharArray()

                Dim split As String() = curLine.Split(delimiter)
                Dim firstWord As String = split(0)
                Dim firstSecondWord As String = ""
                If split.Length > 1 Then
                    firstSecondWord = split(0) + split(1)
                End If

                If curLine.StartsWith("'") OrElse curLine.Trim = "" Then
                    'Comment or whitespace found
                ElseIf firstWord.ToLower.CompareTo(NAMESPACE_STMT) = 0 Then
                    ' Namespace declaration found
                    m_namespaces.Push(CStr(IIf(m_namespaces.Count > 0, ".", "")) & curLine.Substring(curLine.ToLower.IndexOf(NAMESPACE_STMT) + (NAMESPACE_STMT.Length)).Trim)
                ElseIf firstWord.ToLower.CompareTo(END_STMT) = 0 Then
                    tempCurLine = curLine.Replace(" ", "").Replace(vbTab, "").ToLower()
                    If curLine.Replace(" ", "").ToLower.StartsWith(END_NAMESPACE_STMT) OrElse curLine.Replace(vbTab, "").ToLower.StartsWith(END_NAMESPACE_STMT) Then
                        ' End Namespace found
                        If m_namespaces.Count <> 0 Then
                            m_namespaces.Pop()
                        End If
                    ElseIf tempCurLine.StartsWith("endclass") OrElse tempCurLine.StartsWith("endmodule") OrElse tempCurLine.StartsWith("endinterface") OrElse _
                            tempCurLine.StartsWith("endstructure") Then
                        ' End Class found
                        If m_classes.Count <> 0 Then
                            m_classes.Pop()
                        End If
                    ElseIf tempCurLine.StartsWith("endenum") Then
                        ' end enum found
                        If m_enums.Count <> 0 Then
                            m_enums.Pop()
                        End If
                    End If

                Else

                    'Fix for bug#000073 For some reason signature was empty here when it should not have been. 
                    signature = curLine.Trim()

                    Dim className As New Regex("(?i)^(((\<[^\>]>|Public|Private|Protected|Friend|Protected\s+Friend|Shadows|MustInherit|NotInheritable|\s+)+\s+(class|module|structure|interface))|class|module|structure|interface)\s+(?<name>\S*)")
                    Dim match1 As Match
                    Dim match2 As Match
                    match1 = className.Match(curLine)
                    While signature.StartsWith("<")
                        If signature.IndexOf(">") > 0 Then
                            signature = signature.Substring(signature.IndexOf(">") + 1)
                            signature = signature.TrimStart()
                        Else
                            Exit While
                        End If
                    End While
                    match2 = className.Match(signature)
                    If match1.Success Then
                        'encountered new Class
                        m_classes.Push(CStr(IIf(m_classes.Count > 0, "+", ".")) & match1.Result("${name}"))
                    ElseIf (match2.Success) Then
                        m_classes.Push(CStr(IIf(m_classes.Count > 0, "+", ".")) & match2.Result("${name}"))
                    Else
                        Dim enumName As New Regex("(?i)^(((\<[^\>]>|Public|Private|Protected|Friend|Protected\s+Friend|Shadows||\s+)+\s+(enum))|enum)\s+(?<name>\S*)")
                        match1 = enumName.Match(curLine)
                        While signature.StartsWith("<")
                            If signature.IndexOf(">") > 0 Then
                                signature = signature.Substring(signature.IndexOf(">") + 1)
                                signature = signature.TrimStart()
                            Else
                                Exit While
                            End If
                        End While
                        match2 = enumName.Match(signature)
                        If match1.Success Then
                            'encountered new Enum
                            m_enums.Push(CStr(IIf(m_enums.Count > 0, "+", ".")) & match1.Result("${name}"))
                        ElseIf (match2.Success) Then
                            m_enums.Push(CStr(IIf(m_enums.Count > 0, "+", ".")) & match2.Result("${name}"))

                        End If
                    End If
                End If



                If SubLines.Count > 0 Then
                    curLine = SubLines.Item(0)
                    SubLines.RemoveAt(0)
                Else
                    curLine = reader.ReadLine()
                    mLineNumber += 1
                    'curLine = afterColon
                End If

            End While
        Catch exp As Exception
            m_logWriter.WriteLine("Unexpected Error in ParseFile Method" & Environment.NewLine & exp.Message)

        Finally
            reader.Close()
        End Try
    End Sub

#End Region

#Region " ParseProjectFiles "

    ''' <summary>
    ''' A recursive method for visiting all files in a Project, including those in subdirectories
    ''' </summary>
    ''' <param name="projItems"></param>
    Protected Sub ParseProjectFiles(ByVal projItems As ProjectItems)
        Dim SubProjectItems As ProjectItems
        Dim FileName As String

        Dim projItem As ProjectItem
        For Each projItem In projItems
            Try
                SubProjectItems = projItem.ProjectItems
                FileName = DirectCast(projItem.Properties.Item("fullpath").Value, String)
                ' Only attempt to parse VB files
                If FileName.EndsWith(".vb") Then
                    If Not Directory.Exists(FileName) Then
                        Dim CommentTestReader As StreamReader
                        Try
                            CommentTestReader = New StreamReader(FileName, System.Text.Encoding.GetEncoding("Shift_JIS"))
                            If CommentTestReader.ReadToEnd.IndexOf(Configuration.GetInstance().XMLCommentPrefix) <> -1 Then
                                m_logWriter.WriteLine("Parsing file: " & FileName)
                                m_curFileProcessing = FileName
                                ParseFile(New StreamReader(FileName, System.Text.Encoding.GetEncoding("Shift_JIS")))
                            End If
                        Catch Ex As Exception
                            m_logWriter.WriteLine("Error reading: " & FileName & Environment.NewLine & ex.Message)
                        Finally
                            CommentTestReader.Close()
                        End Try
                    End If
                End If
                If SubProjectItems.Count > 0 Then
                    ParseProjectFiles(SubProjectItems)
                End If
                If Not projItem.SubProject Is Nothing Then
                    ParseProjectFiles(projItem.SubProject.ProjectItems)
                End If
            Catch Ex As Exception
                m_logWriter.WriteLine("Error in ParseProjectFiles Method: " & Environment.NewLine & Ex.Message)
            End Try
        Next
        ' Moved this from the ValidateXml method for performance reasons
        If haveTaskList = True Then
            tasklist.TaskItems.ForceItemsToTaskList()
        End If
    End Sub

#End Region

#Region " Write Method "

    ''' <summary>
    ''' The high-level method called after a project builds to kick off the Xml production process
    ''' </summary>
    ''' <param name="projectStr"></param>
    ''' <param name="confStr"></param>
    ''' <param name="platformStr"></param>
    Public Sub Write(ByVal projectStr As String, ByVal confStr As String, ByVal platformStr As String)
        Dim projItemCnt As Integer = 0

        Try
            Me.m_confStr = confStr
            Me.m_platformStr = platformStr
            m_project = FindProject(projectStr)
            If Not m_project Is Nothing Then
                InitializeAssemblyInfoAndXmlWriter()

                m_namespaces = New Stack
                projItemCnt = 1

                If DirectCast(m_project.Properties.Item("rootnamespace").Value, String) <> "" Then
                    m_namespaces.Push(CType(m_project.Properties.Item("rootnamespace").Value, String))
                End If
                Try
                    m_appObj.StatusBar.Progress(True, "Building XML Comment file...", projItemCnt, m_project.ProjectItems.Count)

                Catch ex As Exception

                End Try
                projItemCnt += 1

                ParseProjectFiles(m_project.ProjectItems)

                If Not m_xmlWriter Is Nothing Then
                    ' Close members and doc tags
                    m_xmlWriter.WriteEndElement()
                    m_xmlWriter.WriteEndElement()

                    m_xmlWriter.WriteEndDocument()
                    m_xmlWriter.Close()
                    m_xmlWriter = Nothing

                    m_logWriter.Close()
                    m_logWriter = Nothing

                    If m_intermediatePath <> m_projectOutputPath Then
                        File.Copy(m_intermediatePath & m_assemblyName & ".xml", _
                        m_projectOutputPath & m_assemblyName & ".xml", True)
                    End If

                    File.Copy(m_intermediatePath & m_assemblyName & ".xml", _
                    m_projectPath & m_assemblyName & ".xml", True)
                End If
                m_localNamespaces = Nothing
                m_importedNamespaces = Nothing
                Try
                    m_appObj.StatusBar.Progress(False, "Building XML Commenter file...", 0, 0)
                Catch ex As Exception

                End Try
            End If
        Catch exp As Exception
            m_logWriter.WriteLine("Unexpected Exception in Write Method" & Environment.NewLine & exp.Message)
        Finally
            If Not m_xmlWriter Is Nothing Then
                m_xmlWriter.Close()
                m_xmlWriter = Nothing
            End If
            If Not m_logWriter Is Nothing Then
                m_logWriter.Close()
                m_logWriter = Nothing
            End If
        End Try
    End Sub

    Private Function FindProject(ByVal projectName As String) As Project
        Dim proj As Project
        Dim retProj As Project
        Dim name As String

        ' This is the old for each loop - this caused any solution with a reporting project to lose intellisense and fail
        ' For Each proj In m_appObj.Solution.Projects
        ' The COM object's enumerator (NewEnum?) seems to choke on Reporting projects, 

        ' Yes, this is a 1-based array (remember, COM objects)
        For i As Integer = 1 To m_appObj.Solution.Projects.Count
            proj = m_appObj.Solution.Projects.Item(i)
            ' Try to determine if this is an Enterprise Template project - if so, we need to actually loop through it's projectItems
            ' collection to find the subprojects
            If proj.Kind = "{7D353B21-6E36-11D2-B35A-0000F81F0C06}" Then
                Dim projObj As ProjectItem
                For j As Integer = 1 To proj.ProjectItems.Count
                    projObj = proj.ProjectItems.Item(j)
                    If projObj.Kind = "{EA6618E8-6E24-4528-94BE-6889FE16485C}" Then
                        Try
                            name = projObj.Name
                        Catch ex As Exception
                            name = Nothing
                        End Try
                        If Not name Is Nothing AndAlso (projectName.ToLower.EndsWith(name.ToLower & ".vbproj") Or projectName.ToLower.EndsWith(name.ToLower & ".vbdproj")) Then
                            ' Found a match 
                            Return projObj.SubProject
                        End If
                    End If
                Next
            Else
                ' SW 14/8/04 - we need to make sure it is a project we can handle - database project don't have a valid Name property for example 
                ' and thus cause exceptions to be thrown - making the loop break and any projects after the database project be skipped
                Try
                    name = proj.Name
                Catch ex As Exception
                    name = Nothing
                End Try
                ' JG 11-07-2003: Check for smart device projects, too

                If Not name Is Nothing AndAlso (projectName.ToLower.EndsWith(name.ToLower & ".vbproj") Or projectName.ToLower.EndsWith(name.ToLower & ".vbdproj")) Then
                    Return proj
                End If
            End If
        Next
        ' If we get here, we couldn't find a matching project, so return nothing.
        Return Nothing
    End Function
#End Region

#Region " WriteXML Method "

    ''' <summary>
    ''' This method handles writing an Xml node for one item, either Method, Property, or Class
    ''' </summary>
    ''' <param name="header">xml-comments for the item passed in</param>
    ''' <param name="signature">signature of the item passed in</param>
    Protected Sub WriteXML(ByVal header As StringBuilder, ByVal signature As String)
        Try
            Dim sigMatch As Match = New Regex(VBCommenter.SIGNATURE_MODIFIERS).Match(signature)
            'Could use this for matching headers
            'Match headerMatch = (new Regex(m_HEADER_EXPR)).Match(header);
            Dim paramsMatch As Match
            Dim RegexParams As New Regex(m_PARAMS_EXPR)
            Dim namespaceStr As String = ""
            Dim classStr As String = ""
            Dim enumStr As String = ""
            Dim qualifiedSignature As String = ""
            Dim paramTypes As New ArrayList
            Dim paramNames As New ArrayList
            Dim paramArrays As New ArrayList
            Dim paramByRefs As New ArrayList
            Dim type As String
            Dim name As String
            Dim isShared As Boolean = False

            If Not sigMatch.Success Then
                Exit Sub
            End If

            Dim nmspc As String
            For Each nmspc In m_namespaces
                namespaceStr = nmspc & namespaceStr
            Next

            Dim clss As String
            For Each clss In m_classes
                classStr = clss & classStr
            Next
            classStr &= "."

            Dim en As String
            For Each en In m_enums
                enumStr = en & enumStr
            Next
            enumStr &= "."

            type = sigMatch.Result("${type}").ToLower
            name = sigMatch.Result("${name}")

            ' is this shared - only important for the static constructor
            isShared = (sigMatch.Result("${shared}") = "Shared")

            Dim typePrefix As String = ""

            Select Case type
                ' only do the param search on types that take params
            Case "property", "function", "sub", "event"
                    paramsMatch = RegexParams.Match(sigMatch.Result("${params}"))

                    While paramsMatch.Success
                        paramTypes.Add(paramsMatch.Result("${type}"))
                        paramNames.Add(paramsMatch.Result("${name}"))
                        paramByRefs.Add(paramsMatch.Result("${byref}"))
                        If paramsMatch.Result("${array1}").Trim <> String.Empty Then
                            paramArrays.Add(paramsMatch.Result("${array1}"))
                        ElseIf paramsMatch.Result("${array2}").Trim <> String.Empty Then
                            paramArrays.Add(paramsMatch.Result("${array2}"))
                        Else
                            paramArrays.Add(String.Empty)
                        End If

                        paramsMatch = paramsMatch.NextMatch
                    End While
                Case Else
                    ' do nothing
            End Select
            Dim sigType As Boolean = False
            Select Case type
                Case "property"
                    qualifiedSignature = GetPropertySignature(namespaceStr & classStr & name, paramTypes, paramNames, paramByRefs, paramArrays)
                    typePrefix = "P:"
                Case "function", "sub"
                    If sigMatch.Result("${delegate}").Trim <> String.Empty Then
                        qualifiedSignature = namespaceStr & classStr & name
                        typePrefix = "T:"
                    Else
                        qualifiedSignature = GetMethodSignature(namespaceStr & classStr & name, paramTypes, paramNames, paramByRefs, paramArrays, isShared)
                        typePrefix = "M:"
                    End If
                Case "class", "module", "structure", "interface", "enum"
                    qualifiedSignature = namespaceStr & classStr & name
                    typePrefix = "T:"
                    sigType = True
                Case "event"
                    qualifiedSignature = namespaceStr & classStr & name
                    typePrefix = "E:"
                Case Else
                    ' TODO: Find a better way to do this.  This doesn't really always have to be a field.
                    If enumStr <> "." Then
                        ' don't use "type" because it's "ToLower'ed"
                        qualifiedSignature = namespaceStr & classStr.TrimEnd("."c) & enumStr & sigMatch.Result("${type}") & name
                        typePrefix = "F:"
                    ElseIf classStr <> "." Then
                        qualifiedSignature = namespaceStr & classStr & sigMatch.Result("${type}")
                        ' DTR2 - 2/25/2005
                        ' In VB.NET, a WithEvents "Field" becomes a property once compiled - because it seems there is no way 
                        ' for NDoc or other reflection-based tools to tell the that the emitted property is really a "WithEvents" field,
                        ' VBCommenter needs to emit the documentation as if it were a proeprty
                        Dim isWithEvents As Boolean = (sigMatch.Result("${withEvents}") = "WithEvents")
                        If isWithEvents Then
                            typePrefix = "P:"
                        Else
                            typePrefix = "F:"
                        End If
                    Else
                        Exit Sub
                    End If
                    ' If Type is really the name, replace name with type so we can use it in ProcessCRef
                    name = type
            End Select

            qualifiedSignature = CleanUpSignature(qualifiedSignature)

            ProcessCrefs(header, qualifiedSignature, sigType, namespaceStr, classStr.Substring(0, classStr.Length - 1))
            Dim finalHeader As String = header.ToString()
            Dim inputXml As String = "<member name=""" & typePrefix & qualifiedSignature & """>" & finalHeader & "</member>"

            ValidateXML(inputXml)

            WriteElement(inputXml, signature, finalHeader)

            If typePrefix = "E:" Then
                If paramTypes.Count > 0 Or signature.IndexOf(" As ") < 0 Then
                    ' we've found a VB style Event declaration (i.e. without a delegate), so we'll have to dummy
                    ' up a delegate.  First do the event.
                    qualifiedSignature = namespaceStr & classStr & name & "EventHandler"
                    inputXml.Replace("<member name=""E:" & qualifiedSignature, "<member name=""T:" & qualifiedSignature & "EventHandler")
                    WriteElement(inputXml, signature, finalHeader)
                End If
            End If

        Catch exp As Exception
            m_logWriter.WriteLine("Unexpected error in WriteXML() Method" & Environment.NewLine & exp.Message)
        End Try
    End Sub
#End Region

#Region "Process CREF attributes"

    Private Sub ProcessCrefs(ByRef header As StringBuilder, ByVal qualifiedSignature As String, ByVal sigIsClassType As Boolean, ByVal namespaceStr As String, ByVal classStr As String)
        Dim tmpHeader As String = header.ToString()
        Dim crefMatches As MatchCollection = New Regex(m_CREF_EXPR).Matches(tmpHeader)
        Dim replacements As System.Collections.Specialized.StringCollection = New System.Collections.Specialized.StringCollection
        Dim userSuppliedTester As Regex = New Regex(m_USER_SUPPLIED_MARKER_EXPR)
        For Each crefMatch As Match In crefMatches
            If crefMatch.Success Then
                ' First, see if the CREF is just a type reference
                Dim crefTxt As String = crefMatch.Result("${cref}")
                ' Avoid doing the same replacement twice in the same header, which could cause problems
                If Not replacements.Contains(crefTxt) Then
                    replacements.Add(crefTxt)
                    ' we will assume that the user knows what s/he is doing if they give us a cref with a typed reference
                    If Not userSuppliedTester.IsMatch(crefTxt) Then

                        Dim memberName As String = String.Empty
                        Dim t As System.Type = DeriveType(crefTxt, memberName)
                        If t Is Nothing Then
                            t = DeriveType(qualifiedSignature + "." + crefTxt, memberName)
                        End If
                        If t Is Nothing Then
                            If haveTaskList Then
                                tasklist.TaskItems.Add(TaskListCategory, "Error", "Invalid cref - " & crefTxt, vsTaskPriority.vsTaskPriorityMedium, vsTaskIcon.vsTaskIconCompile, False, Me.m_curFileProcessing, mCommentStartLineNumber, True, False)
                            End If
                        Else
                            If memberName = String.Empty Then
                                ' The DeriveType returns something and memberName is still String.Empty, then the cref is a type reference.
                                header = header.Replace("cref=""" & crefTxt & """", "cref=""T:" & t.FullName.Replace("+"c, "."c) & """")
                            Else
                                ' See if we found a type that is a substring of our crefTxt - if 
                                If memberName.ToLower = "new" Then
                                    memberName = ".ctor"
                                End If
                                Dim memberInfo() As MemberInfo = t.FindMembers(MemberTypes.All, BindingFlags.Public Or BindingFlags.Static Or BindingFlags.NonPublic Or BindingFlags.IgnoreCase Or BindingFlags.Instance, t.FilterNameIgnoreCase, memberName)
                                For Each member As MemberInfo In memberInfo
                                    If TypeOf member Is ConstructorInfo Then
                                        Dim ctorInfo As ConstructorInfo = CType(member, ConstructorInfo)
                                        Dim params() As ParameterInfo = ctorInfo.GetParameters()
                                        Dim paramStr As String = DeriveParamString(params)
                                        header = header.Replace("cref=""" & crefTxt & """", "cref=""M:" & ctorInfo.DeclaringType.FullName.Replace("+"c, "."c) & ".#ctor" & paramStr + """")
                                        Exit For
                                    End If
                                    If TypeOf member Is MethodInfo Then
                                        Dim methInfo As MethodInfo = CType(member, MethodInfo)
                                        Dim params() As ParameterInfo = methInfo.GetParameters()
                                        Dim paramStr As String = DeriveParamString(params)
                                        header = header.Replace("cref=""" & crefTxt & """", "cref=""M:" & methInfo.DeclaringType.FullName.Replace("+"c, "."c) & "." & methInfo.Name & paramStr & """")
                                        Exit For
                                    ElseIf TypeOf member Is PropertyInfo Then
                                        Dim propInfo As PropertyInfo = CType(member, PropertyInfo)
                                        Dim meth As MethodInfo
                                        Dim methIsSet As Boolean
                                        ' Changed this logic to call the Get*Methods with the True parameter, as we may need to reference Non-Public properties
                                        If propInfo.CanRead Then
                                            meth = propInfo.GetGetMethod(True)
                                            methIsSet = False
                                        Else
                                            meth = propInfo.GetSetMethod(True)
                                            methIsSet = True
                                        End If
                                        If Not meth Is Nothing Then
                                            Dim params() As ParameterInfo = meth.GetParameters()
                                            Dim paramStr As String = String.Empty
                                            If params.Length > 0 Then
                                                If methIsSet And params.Length > 1 Then
                                                    Dim newParams(params.Length - 2) As ParameterInfo
                                                    params.Copy(params, 1, newParams, 0, params.Length - 1)
                                                    params = newParams
                                                End If
                                                paramStr = DeriveParamString(params)
                                            End If
                                            header = header.Replace("cref=""" & crefTxt & """", "cref=""P:" & propInfo.DeclaringType.FullName.Replace("+"c, "."c) & "." & propInfo.Name & paramStr & """")
                                            Exit For
                                        End If
                                    ElseIf TypeOf member Is FieldInfo Then
                                        header = header.Replace("cref=""" & crefTxt & """", "cref=""F:" & member.DeclaringType.FullName.Replace("+"c, "."c) & "." & member.Name & """")
                                        Exit For
                                    ElseIf TypeOf member Is EventInfo Then
                                        header = header.Replace("cref=""" & crefTxt & """", "cref=""E:" & member.DeclaringType.FullName.Replace("+"c, "."c) & "." & member.Name & """")
                                        Exit For
                                    ElseIf TypeOf member Is System.Type Then
                                        ' Nested type declaration
                                        header = header.Replace("cref=""" & crefTxt & """", "cref=""T:" & member.DeclaringType.FullName.Replace("+"c, "."c) & "." & member.Name & """")
                                        Exit For
                                    End If
                                Next
                            End If
                        End If
                    End If
                End If
            End If
        Next
    End Sub

    Private Function DeriveType(ByVal typeStr As String, ByRef memberName As String) As System.Type
        Dim t As System.Type
        Dim origTypeStr As String = typeStr
        Dim dotIndx As Integer
        t = FindType(typeStr)
        If t Is Nothing Then
            Dim mName As String
            ' We need to start by trying to take off the last (.*) group in the string and deriving its type, as otherwise we may match a class name in another namespace
            dotIndx = origTypeStr.LastIndexOf(".")
            If dotIndx >= 0 Then
                mName = origTypeStr.Substring(dotIndx + 1)
                typeStr = origTypeStr.Substring(0, dotIndx)
                t = DeriveType(typeStr, mName)
                If Not t Is Nothing Then
                    memberName = mName
                    Return t
                End If
            End If
        End If
        ' If we didn't find a match that way, 

        Return t
    End Function

#End Region

#Region "Derive Param String (for ProcessCrefs)"
    Private Function DeriveParamString(ByVal params() As ParameterInfo) As String
        Dim paramStr As String = String.Empty
        For Each param As ParameterInfo In params
            paramStr &= GetParamFullName(param) & ","
        Next
        If params.Length > 0 Then
            paramStr = "(" & paramStr.Substring(0, paramStr.Length - 1) & ")"
        Else
            paramStr = ""
        End If
        Return paramStr
    End Function

#End Region

    '''<summary>Validate the XML fragment using an XML Schema definition</summary>
    Private Sub ValidateXML(ByVal xmlFrag As String)
        Dim reader As System.Xml.XmlValidatingReader = Nothing

        Try

            'Create the XmlParserContext.
            Dim context As New System.Xml.XmlParserContext(Nothing, Nothing, "", System.Xml.XmlSpace.None)

            'Implement the reader. 
            reader = New System.Xml.XmlValidatingReader(xmlFrag, System.Xml.XmlNodeType.Element, context)

            'Set the schema type and add the schema to the reader.
            reader.ValidationType = System.Xml.ValidationType.Schema
            reader.Schemas.Add(myschema)

            'Add the handler to raise the validation event.
            AddHandler reader.ValidationEventHandler, AddressOf ValidationEventHandle

            ' read lines 
            While reader.Read()
            End While

            ' we added items to the task list with flushitem=false, so now
            ' we need to tell it to update
        Catch XmlExp As System.Xml.XmlException
            Me.LogValidationError(XmlExp)
        Catch XmlSchExp As System.Xml.schema.XmlSchemaException
            Me.LogValidationError(XmlSchExp)
        Catch GenExp As Exception
            'Console.WriteLine(GenExp.Message)

        Finally
            reader.Close()

        End Try
    End Sub

    Public Sub ValidationEventHandle(ByVal sender As Object, ByVal args As System.Xml.Schema.ValidationEventArgs)

        Me.LogValidationError(args.Exception)

    End Sub 'ValidationEventHandle

    Public Sub LogValidationError(ByVal ex As System.Xml.Schema.XmlSchemaException)
        If haveTaskList Then
            Dim msg As String = ex.Message
            tasklist.TaskItems.Add(TaskListCategory, "Error", msg.Substring(0, msg.LastIndexOf("."c, ex.Message.Length - 2)), vsTaskPriority.vsTaskPriorityMedium, vsTaskIcon.vsTaskIconCompile, False, Me.m_curFileProcessing, mCommentStartLineNumber + ex.LineNumber - 1, True, False)
        End If
    End Sub

    Public Sub LogValidationError(ByVal ex As System.Xml.XmlException)
        If haveTaskList Then
            Dim msg As String = ex.Message
            tasklist.TaskItems.Add(TaskListCategory, "Error", msg.Substring(0, msg.LastIndexOf("."c, ex.Message.Length - 2)), vsTaskPriority.vsTaskPriorityMedium, vsTaskIcon.vsTaskIconCompile, False, Me.m_curFileProcessing, mCommentStartLineNumber + ex.LineNumber - 1, True, False)
        End If
    End Sub

#Region " WriteElement Method"

    Protected Sub WriteElement(ByVal inputXml As String, ByVal signature As String, ByVal header As String)
        Dim xmlReader As XmlTextReader = Nothing
        Try

            xmlReader = New XmlTextReader(New StringReader(inputXml))
            xmlReader.WhitespaceHandling = WhitespaceHandling.All
            Dim hitEmptyElement As Boolean

            While xmlReader.Read()
                Select Case xmlReader.NodeType
                    Case XmlNodeType.Element
                        ' process <include> tags
                        If xmlReader.LocalName.ToLower = "include" Then
                            Dim xmlInclude As New Xml.XmlDocument
                            'check first for absolute path that works
                            If File.Exists(xmlReader.GetAttribute("file")) Then
                                xmlInclude.Load(xmlReader.GetAttribute("file"))
                            Else
                                xmlInclude.Load(Path.GetDirectoryName(m_curFileProcessing) & "\" & xmlReader.GetAttribute("file"))
                            End If
                            Dim xmlContent As XmlNodeList = xmlInclude.SelectNodes(xmlReader.GetAttribute("path"))
                            Dim xmlNode As XmlNode
                            For Each xmlNode In xmlContent
                                m_xmlWriter.WriteRaw(xmlNode.OuterXml)
                            Next
                        Else
                            m_xmlWriter.WriteStartElement(xmlReader.Name)

                            If xmlReader.IsEmptyElement Then
                                hitEmptyElement = True
                            End If

                            If xmlReader.HasAttributes Then m_xmlWriter.WriteAttributes(xmlReader, False)

                            If hitEmptyElement Then
                                m_xmlWriter.WriteEndElement()
                                hitEmptyElement = False
                            End If
                        End If
                    Case XmlNodeType.Text
                        m_xmlWriter.WriteString(xmlReader.Value)

                    Case XmlNodeType.EndElement
                        m_xmlWriter.WriteEndElement()
                    Case XmlNodeType.EntityReference
                        m_xmlWriter.WriteEntityRef(xmlReader.Name)
                    Case (XmlNodeType.Whitespace)
                        m_xmlWriter.WriteWhitespace(xmlReader.Value)

                End Select
            End While
        Catch exp As System.Xml.XmlException
            ' Hit some malformed xml-comments
            Dim errorMessage As String = exp.Message & Environment.NewLine & Environment.NewLine & "Filename: " & m_curFileProcessing & Environment.NewLine & Environment.NewLine & "Signature: " & signature & Environment.NewLine & Environment.NewLine & "Header: " & header & Environment.NewLine & Environment.NewLine & "Skipping and moving on...."
            m_logWriter.WriteLine(errorMessage)
        Finally
            xmlReader.Close()
            m_xmlWriter.Flush()
        End Try
    End Sub
#End Region

#Region " InitializeAssemblyInfoAndXmlWriter Method "

    ''' <summary>
    ''' Verify that the assembly is loaded, XmlWriter is properly initialized, assemblies types are loaded, and logWriter is prepared
    ''' </summary>
    Protected Sub InitializeAssemblyInfoAndXmlWriter()

        m_projectPath = DirectCast(m_project.Properties.Item("fullpath").Value, String)

        If m_project.ConfigurationManager.Item(m_confStr, m_platformStr).Properties Is Nothing Then
            m_intermediatePath = m_projectPath & "obj\Debug\"
            m_projectOutputPath = m_projectPath & "bin\Debug\"
        Else
            'WiebeHordijk: fixed bug: project path was left out
            m_intermediatePath = m_projectPath & DirectCast(m_project.ConfigurationManager.Item(m_confStr, m_platformStr).Properties.Item("intermediatepath").Value, String)
            m_projectOutputPath = m_projectPath & DirectCast(m_project.ConfigurationManager.Item(m_confStr, m_platformStr).Properties.Item("outputpath").Value, String)
        End If

        m_assemblyName = DirectCast(m_project.Properties.Item("assemblyname").Value, String)

        ' Hook into the AssemblyResolve event of the AppDomain to make sure we can try to load dependent assemblies
        AddHandler AppDomain.CurrentDomain.AssemblyResolve, AddressOf AssemblyResolveEventHandler

        ' Load the assembly into memory
        LoadAssembly()

        ' Create the xml output file if we haven't yet
        If m_xmlWriter Is Nothing Then
            m_xmlWriter = New XmlTextWriter(m_intermediatePath & m_assemblyName & ".xml", System.Text.Encoding.UTF8)
            m_xmlWriter.Formatting = Formatting.Indented
            m_xmlWriter.WriteStartDocument()
            m_xmlWriter.WriteStartElement("doc")
            m_xmlWriter.WriteStartElement("assembly")
            m_xmlWriter.WriteElementString("name", DirectCast(m_project.Properties.Item("assemblyname").Value, String))
            m_xmlWriter.WriteElementString("version", m_assembly.GetName().Version.ToString)
            m_xmlWriter.WriteElementString("fullname", m_assembly.FullName)
            m_xmlWriter.WriteEndElement()

            m_xmlWriter.WriteStartElement("members")
        End If

        ' Create the log output file if we haven't yet
        If m_logWriter Is Nothing Then
            m_logWriter = New StreamWriter(m_intermediatePath & "VBCommenterLog.txt")

#If DEBUG Then
            m_logWriter.AutoFlush = True
#End If
        End If

        ' Load all top level classes defined in the assembly if we haven't yet
        If m_topLevelTypes Is Nothing Then
            Try
                m_topLevelTypes = m_assembly.GetTypes
            Catch exp As ReflectionTypeLoadException
                ' If some of the types couldn't be returned, get the successful ones out of the Exception
                m_topLevelTypes = exp.Types
            Catch exp As Exception
                m_logWriter.WriteLine("Unexpected error loading topLevelTypes:")
                m_logWriter.WriteLine(exp.Message)
            End Try
        End If

        ' Load all nested classes if we haven' yet
        If m_allTypes Is Nothing Then
            m_allTypes = New ArrayList
            GetTypes(m_topLevelTypes, m_allTypes)
        End If

    End Sub

#End Region

#Region " GetTypes Method "

    ''' <summary>
    ''' Builds an array of all types in an assembly.
    ''' Note: this method is currently NOT untilized by this addin.
    ''' </summary>
    ''' <param name="types"></param>
    ''' <param name="allTypes"></param>
    Protected Sub GetTypes(ByVal types() As Type, ByVal allTypes As ArrayList)
        If types.Length = 0 Then
            Exit Sub
        End If

        Dim Type As Type
        For Each Type In types
            If Not Type Is Nothing Then
                allTypes.Add(Type)
            End If

            ' start new code
            Dim nestedTypes() As Type
            Try
                nestedTypes = Type.GetNestedTypes
            Catch exp As ReflectionTypeLoadException
                nestedTypes = exp.Types
            Catch exp As Exception
                Console.WriteLine(exp.ToString)
            End Try
            ' end new code

            GetTypes(nestedTypes, allTypes)
        Next
    End Sub

#End Region

#Region " FindType Method "

    ''' <summary>
    ''' For a given typeName, returns the Type objectfrom the project's namespaces
    ''' </summary>
    ''' <param name="typeName">name of type trying to find a match for</param>
    ''' <param name="assembly">namespaces in the assembly</param>
    ''' <returns></returns>
    Protected Overloads Function FindType(ByVal typeName As String, ByVal assem As NamespaceCollection) As Type
        Dim type As Type
        Dim oNSEnum As NamespaceEnumerator
        Dim sNamespaceName As String
        Dim oAssembly As [Assembly]
        Dim aAssemblies() As [Assembly]
        Dim sCurrentNamespace As String

        ' First, try to short-circuit the process and see if it is a fully qualified typename in a default namespace
        type = System.Type.GetType(typeName)

        If type Is Nothing Then
            ' If the typeName does have any qualification, try all combinations 
            ' of types and namespaces to generate a type
            '
            ' Example:
            '		imports System;
            '		...
            '		public sub foo(ByRef param1 As Windows.Forms.Form)
            '		end sub
            Dim dotNdx As Integer = 0
            Dim partialNamespace As String = typeName.Substring(0, dotNdx)
            Dim newType As String = typeName.Substring(dotNdx)
            While dotNdx <> -1 AndAlso type Is Nothing
                oNSEnum = assem.GetNamespaceEnumerator
                While oNSEnum.MoveNext And (type Is Nothing)

                    sNamespaceName = oNSEnum.NamespaceName
                    If sNamespaceName.EndsWith(partialNamespace) Then
                        type = assem.GetCachedType(sNamespaceName, typeName)
                        If Not type Is Nothing Then
                            Return type
                        End If
                        ' sCurrentNamespace = sNamespaceName & "." & typeName.Substring(0, dotNdx)

                        If assem.Exists(sNamespaceName) Then

                            aAssemblies = assem.GetAssembliesForNamespace(sNamespaceName)
                            If Not aAssemblies Is Nothing Then
                                For Each oAssembly In aAssemblies
                                    type = oAssembly.GetType(sNamespaceName + "." + newType, False, True)
                                    If Not type Is Nothing Then
                                        Exit For
                                    End If
                                    ' Check to see if the type happens to be a nested type 
                                    Dim nestedType As String = newType
                                    While type Is Nothing
                                        Dim dotNdx2 As Integer = nestedType.LastIndexOf(".")
                                        ' Try to see if this is a nested type
                                        ' Working from right->left, replace a "." with a "+"
                                        dotNdx2 = nestedType.LastIndexOf("."c)
                                        If dotNdx2 >= 0 Then
                                            nestedType = nestedType.Substring(0, dotNdx2) + "+" + nestedType.Substring(dotNdx2 + 1)
                                            type = oAssembly.GetType(sNamespaceName + "." + nestedType, False, True)
                                        Else
                                            Exit While
                                        End If
                                    End While
                                Next
                            End If
                        End If
                    End If
                End While

                dotNdx = typeName.IndexOf(".", dotNdx + 1)
                If dotNdx <> -1 Then
                    partialNamespace = typeName.Substring(0, dotNdx)
                    newType = typeName.Substring(dotNdx + 1)
                End If
            End While
        End If
        assem.CacheMatch(sNamespaceName, typeName, type)
        Return type
    End Function

#End Region

#Region " GetMethodSignature Method "
    ''' <summary>
    ''' Get the Signature of a Method, which will eventually be in the Xml file in the member tag
    ''' </summary>
    ''' <param name="sigName">signature of the method passed in</param>
    ''' <param name="paramTypes">an Array of parameter types for the method signature passed in</param>
    ''' <param name="paramNames">an Array of parameter names for the method signature passed in</param>
    ''' <returns></returns>
    Protected Function GetMethodSignature(ByVal sigName As String, ByVal paramTypes As ArrayList, ByVal paramNames As ArrayList, _
    ByVal paramByRefs As ArrayList, ByVal paramArrays As ArrayList, ByVal isShared As Boolean) As String
        Dim fullSignature As String
        Dim sigMatches As New ArrayList
        Dim fullTypeName As String
        Dim methodName As String
        Dim type As Type

        Try
            If sigName.LastIndexOf(".") > 0 Then
                fullTypeName = sigName.Substring(0, sigName.LastIndexOf("."))
                methodName = sigName.Substring(sigName.LastIndexOf(".") + 1)
            Else
                fullTypeName = sigName
                methodName = sigName
            End If
            Try
                type = m_assembly.GetType(fullTypeName, True)
            Catch e As Exception
                Console.WriteLine(e.Message)

                'Couldn't find a match in the assembly, so build signature string with what we have from the file
                Return GetSignatureFromNamespaces(sigName, paramTypes, paramByRefs, paramArrays)
            End Try

            Dim sigAndParamMatches As New ArrayList
            Dim paramType As Type

            'If the signature is a constructor, store all the constructors matching the name
            If sigName.ToLower.EndsWith(m_CONSTRUCTOR) OrElse sigName.ToLower.EndsWith(XML_CONSTRUCTOR) Then
                If sigName.ToLower.EndsWith(m_CONSTRUCTOR) Then
                    If isShared Then
                        sigName = sigName.Substring(0, sigName.LastIndexOf(".")) & sigName.Substring(sigName.LastIndexOf(".")).ToLower.Replace(m_CONSTRUCTOR, XML_STATIC_CONSTRUCTOR)
                    Else
                        sigName = sigName.Substring(0, sigName.LastIndexOf(".")) & sigName.Substring(sigName.LastIndexOf(".")).ToLower.Replace(m_CONSTRUCTOR, XML_CONSTRUCTOR)
                    End If

                End If

                Dim constructorInfo As ConstructorInfo
                For Each constructorInfo In type.GetConstructors(BindingFlags.Public Or BindingFlags.NonPublic Or BindingFlags.Instance Or BindingFlags.DeclaredOnly Or BindingFlags.Static)
                    If isShared And constructorInfo.IsStatic Then
                        sigMatches.Add(constructorInfo)
                    ElseIf constructorInfo.IsStatic = False Then
                        sigMatches.Add(constructorInfo)
                    End If
                Next
            Else
                'We have a Method (not a constructor). Add all the methods that have the same name as the signature
                Dim methodInfo As MethodInfo
                For Each methodInfo In type.GetMethods(BindingFlags.Public Or BindingFlags.NonPublic Or BindingFlags.Instance Or BindingFlags.DeclaredOnly Or BindingFlags.Static)
                    If methodInfo.Name.ToLower = methodName.ToLower Then
                        sigMatches.Add(methodInfo)
                    End If
                Next
            End If

            Dim paramInfo As ParameterInfo

            'Only one method or contructor found with the given name
            If sigMatches.Count = 1 Then
                fullSignature = sigName & "("
                For Each paramInfo In DirectCast(sigMatches(0), MethodBase).GetParameters
                    fullSignature &= GetParamFullName(paramInfo) & ","
                Next

                Return fullSignature.Trim(",".ToCharArray) & ")"
            ElseIf sigMatches.Count > 1 Then
                'Overloaded method or constructor found
                Dim MethodBase As MethodBase
                For Each MethodBase In sigMatches
                    ' Check for same number of params
                    If MethodBase.GetParameters().Length = paramNames.Count Then
                        'Check for same name of params
                        Dim ndx As Integer
                        For ndx = 0 To paramNames.Count - 1
                            If MethodBase.GetParameters()(ndx).Name.ToLower <> DirectCast(paramNames.Item(ndx), String).ToLower Then
                                ' (not MethodBase.GetParameters()(ndx).Name.ToLower.EndsWith(DirectCast(paramNames.Item(ndx), String).ToLower()))
                                Exit For
                                'End If
                            End If
                        Next

                        If (ndx >= paramNames.Count) Then
                            sigAndParamMatches.Add(MethodBase)
                        End If
                    End If
                Next

                If sigAndParamMatches.Count = 1 Then
                    fullSignature = sigName & "("
                    For Each paramInfo In CType(sigAndParamMatches.Item(0), MethodBase).GetParameters
                        fullSignature &= GetParamFullName(paramInfo) & ","
                    Next

                    Return fullSignature.Trim(",".ToCharArray()) & ")"
                ElseIf sigAndParamMatches.Count > 1 Then
                    'Overloaded methods/constructors with same number and name of params found
                    Return GetSignatureFromNamespaces(sigName, paramTypes, paramByRefs, paramArrays)
                End If
            End If
        Catch
            'Ignore it, then below we'll write what we can
        End Try
        'Couldn't find a match in the assembly, so build signature string with what we have from the file
        fullSignature = sigName
        If paramTypes.Count > 0 Then
            fullSignature &= "("
            Dim p As String
            For Each p In paramTypes
                fullSignature &= p & ","
            Next
            fullSignature = fullSignature.Substring(0, fullSignature.Length - 1)
            fullSignature &= ")"
        End If

        Return fullSignature
    End Function

#End Region

#Region " GetParamFullName Method "
    Private Function GetParamFullName(ByVal parameter As ParameterInfo) As String
        Dim parameterName As String = parameter.ParameterType.FullName

        parameterName = Regex.Replace(parameterName, ",", ",0:")
        parameterName = Regex.Replace(parameterName, "\[,", "[0:,")

        ' XML Documentation file appends a "@" to reference and out types, not a "&"
        Return parameterName.Replace("&", "@").Replace("+", ".")
    End Function
#End Region

#Region " GetPropertySignature Method "

    ''' <summary>
    ''' Return the full signature for the Property passed in
    ''' </summary>
    ''' <param name="sigName">signature of the Property passed in</param>
    ''' <param name="paramTypes">an Array of parameter types for the Property signature passed in</param>
    ''' <param name="paramNames">an Array of parameter names for the Property signature passed in</param>
    ''' <returns></returns>
    Protected Function GetPropertySignature(ByVal sigName As String, ByVal paramTypes As ArrayList, ByVal paramNames As ArrayList, _
    ByVal paramByRefs As ArrayList, ByVal paramArrays As ArrayList) As String
        Dim fullSignature As String
        Dim sigMatches As New ArrayList
        Dim fullTypeName As String
        Dim propertyName As String
        Dim type As Type
        Dim paramType As Type

        ' don't use this way, it doesn't get correct types for indexers
        'fullSignature = sigName
        'If paramTypes.Count > 0 Then
        '    fullSignature &= "("
        '    For Each p As String In paramTypes
        '        fullSignature &= p & ","
        '    Next
        '    fullSignature = fullSignature.Substring(0, fullSignature.Length - 1)
        '    fullSignature &= ")"
        'End If

        'Return fullSignature

        If paramTypes.Count > 0 Then
            ' An alternative way of getting the signature from the project's assembly, but with slower performance:
            Try
                If sigName.LastIndexOf(".") > 0 Then
                    fullTypeName = sigName.Substring(0, sigName.LastIndexOf("."))
                    propertyName = sigName.Substring(sigName.LastIndexOf(".") + 1)
                Else
                    fullTypeName = sigName
                    propertyName = sigName
                End If

                Try
                    type = m_assembly.GetType(fullTypeName, True)
                Catch e As Exception
                    Console.WriteLine(e.Message)

                    'Couldn't find a match in the assembly, so build signature string with what we have from the file
                    Return GetSignatureFromNamespaces(sigName, paramTypes, paramByRefs, paramArrays)
                End Try


                Dim sigAndParamMatches As New ArrayList

                Dim propertyInfo As PropertyInfo
                For Each propertyInfo In type.GetProperties(BindingFlags.Public Or BindingFlags.NonPublic Or BindingFlags.Instance Or BindingFlags.DeclaredOnly Or BindingFlags.Static)
                    If propertyInfo.Name.ToLower = propertyName.ToLower Then
                        sigMatches.Add(propertyInfo)
                    End If
                Next

                'Only one property found with the given name
                If sigMatches.Count = 1 Then
                    fullSignature = sigName & "("
                    Dim paramInfo As ParameterInfo
                    For Each paramInfo In DirectCast(sigMatches(0), PropertyInfo).GetIndexParameters
                        fullSignature &= paramInfo.ParameterType.FullName & ","
                    Next

                    Return fullSignature.Trim(",".ToCharArray) & ")"
                ElseIf sigMatches.Count > 1 Then
                    'Overloaded property or constructor found
                    For Each propertyInfo In sigMatches
                        ' Check for same number of params
                        If propertyInfo.GetIndexParameters().Length = paramNames.Count Then
                            'Check for same name of params
                            Dim ndx As Integer
                            For ndx = 0 To paramNames.Count - 1
                                If propertyInfo.GetIndexParameters()(ndx).Name.ToLower <> DirectCast(paramNames.Item(ndx), String).ToLower Then
                                    Exit For
                                End If
                            Next

                            If (ndx >= paramNames.Count) Then
                                sigAndParamMatches.Add(propertyInfo)
                            End If
                        End If
                    Next

                    If sigAndParamMatches.Count = 1 Then
                        fullSignature = sigName & "("
                        Dim paramInfo As ParameterInfo
                        For Each paramInfo In CType(sigAndParamMatches.Item(0), PropertyInfo).GetIndexParameters
                            fullSignature &= paramInfo.ParameterType.FullName & ","
                        Next

                        Return fullSignature.Trim(",".ToCharArray()) & ")"
                    ElseIf sigAndParamMatches.Count > 1 Then
                        'Overloaded properties with same number and name of params found
                        Return GetSignatureFromNamespaces(sigName, paramTypes, paramByRefs, paramArrays)
                    End If
                End If
            Catch
                'Ignore it, then below we'll write what we can
            End Try
            'Couldn't find a match in the assembly, so build signature string with what we have from the file
            fullSignature = sigName
            If paramTypes.Count > 0 Then
                fullSignature &= "("
                Dim p As String
                For Each p In paramTypes
                    fullSignature &= p & ","
                Next
                fullSignature = fullSignature.Substring(0, fullSignature.Length - 1)
                fullSignature &= ")"
            End If

            Return fullSignature
        Else
            Return sigName
        End If
    End Function

#End Region

#Region " GetSignatureFromNamespaces Method "
    Protected Function GetSignatureFromNamespaces(ByVal sigName As String, ByVal paramTypes As ArrayList, _
    ByVal paramByRefs As ArrayList, ByVal paramArrays As ArrayList) As String
        Dim fullSignature As String
        Dim paramType As Type
        Dim index As Integer = 0

        Dim ndx As Integer
        For ndx = 0 To paramTypes.Count - 1
            If m_VBPrimitivesToConvert.ContainsKey(DirectCast(paramTypes.Item(ndx), String).ToLower) Then
                paramTypes.Item(ndx) = DirectCast(m_VBPrimitivesToConvert.Item(DirectCast(paramTypes.Item(ndx), String).ToLower), String)
            End If
        Next

        InitNamespaceTables()

        fullSignature = sigName & "("
        Dim typeStr As String
        For Each typeStr In paramTypes
            fullSignature &= GetTypeFromNamespaces(typeStr)

            If CStr(paramArrays(index)) <> String.Empty Then
                fullSignature &= CStr(paramArrays(index))
            End If
            If CStr(paramByRefs(index)) <> String.Empty Then
                fullSignature &= "@"
            End If
            fullSignature &= ","
            index += 1
        Next
        Return fullSignature.Trim(",".ToCharArray()) & ")"

    End Function
#End Region

    Private Sub InitNamespaceTables()

        If m_localNamespaces Is Nothing Then
            m_localNamespaces = New NamespaceCollection
            m_importedNamespaces = New NamespaceCollection

            'Load local namespaces
            LoadNamespaces(m_assembly, m_localNamespaces, False)

            'Collect all the namespaces in the project
            Dim assemblyName As AssemblyName
            For Each assemblyName In m_assembly.GetReferencedAssemblies
                Dim asmbly As [Assembly]
                'try Load with partial name first
                asmbly = [Assembly].LoadWithPartialName(assemblyName.Name)
                If asmbly Is Nothing Then
                    'that didn't work, is there a fullname?
                    If Not assemblyName.FullName Is Nothing Then
                        asmbly = [Assembly].Load(assemblyName.FullName)
                    End If
                    If asmbly Is Nothing Then
                        'ah well, we did our best
                        Dim msg As String = "Unable to load assembly for [" & assemblyName.Name & "]"
                        'throw this exception which will be caught by the WriteXML method
                        Throw New Exception(msg)
                    End If
                End If
                LoadNamespaces(asmbly, m_importedNamespaces, True)
            Next
        End If

    End Sub

#Region " GetTypeFromNamespaces "
    Private Function GetTypeFromNamespaces(ByVal typeStr As String, Optional ByVal emptyIfFails As Boolean = False) As String
        Dim paramType As Type = FindType(typeStr)

        If Not paramType Is Nothing Then
            Return paramType.Namespace & "." & paramType.Name
        Else
            If emptyIfFails Then
                Return String.Empty
            End If
            Return typeStr
        End If
    End Function
#End Region

    Private Overloads Function FindType(ByVal typeStr As String) As System.Type
        InitNamespaceTables()
        Dim paramType As Type
        paramType = FindType(typeStr, m_localNamespaces)

        If paramType Is Nothing Then
            paramType = FindType(typeStr, m_importedNamespaces)
        End If
        Return paramType
    End Function

#Region " CleanUpSignature Method "

    Protected Function CleanUpSignature(ByVal signature As String) As String
        'Classes nested in other classes will have a "+" that we need to replace with "."
        signature = signature.Replace("+", ".")

        'If the signature is obtained via reflection, all byref params will have
        '&'s in front of them
        signature = signature.Replace("&", "")

        'IntelliSense currently won't read constructors w/o params unless they
        'have no parens.
        If signature.IndexOf(XML_CONSTRUCTOR & "()") <> -1 Then
            signature = signature.Replace(XML_CONSTRUCTOR & "()", XML_CONSTRUCTOR)
        End If

        ' no parens are necessary for all methods w/o params
        If signature.EndsWith("()") Then
            signature = signature.Substring(0, signature.Length - 2)
        End If
        Return signature
    End Function

#End Region

#Region " LoadNamespaces Method "

    ''' <summary>
    ''' Load all of the namespaces in the assembly for this project
    ''' </summary>
    ''' <param name="assembly">project assembly</param>
    ''' <param name="assemblyNamespaces">Hashtable to load with this assembly's namespaces</param>
    Protected Sub LoadNamespaces(ByVal assem As [Assembly], ByVal assemblyNamespaces As NamespaceCollection, ByVal exportedTypeNamespacesOnly As Boolean)
        Dim types As New ArrayList
        Dim type As Type

        Try

            If exportedTypeNamespacesOnly Then
                GetTypes(assem.GetExportedTypes(), types)
            Else
                GetTypes(assem.GetTypes(), types)
            End If

            For Each type In types

                If Not type Is Nothing AndAlso Not type.Namespace Is Nothing Then

                    If Not assemblyNamespaces.Exists(type.Namespace, assem) Then
                        assemblyNamespaces.Add(type.Namespace, assem)
                    End If

                ElseIf type Is Nothing Then
                    m_logWriter.WriteLine("Encountered null type in types array")
                End If

            Next

        Catch exp As ArgumentNullException
            m_logWriter.WriteLine("Error in LoadNamespaces Method" & Environment.NewLine & exp.Message)
        End Try
    End Sub

#End Region

#Region " LoadFile Method "

    ''' <summary>
    ''' Load a file into a byte array
    ''' </summary>
    ''' <param name="filename">file to load</param>
    ''' <returns>the file, in the form of a byte array</returns>
    Protected Shared Function LoadFile(ByVal filename As String) As Byte()
        Dim fs As New FileStream(filename, FileMode.Open, FileAccess.Read)
        Dim buffer(CType(fs.Length, Integer)) As Byte
        fs.Read(buffer, 0, buffer.Length)
        fs.Close()

        Return buffer
    End Function

#End Region

#Region " LoadAssembly Method "

    Protected Sub LoadAssembly()
        Dim path As String
        Dim assemBytes() As Byte
        Try
            path = m_intermediatePath & DirectCast(m_project.Properties.Item("outputfilename").Value, String)
            assemBytes = LoadFile(path)
        Catch
            m_intermediatePath = m_projectOutputPath
            path = m_intermediatePath & DirectCast(m_project.Properties.Item("outputfilename").Value, String)
            assemBytes = LoadFile(path)
        End Try

        'm_assembly = [Assembly].LoadFile(path)
        m_assembly = [Assembly].Load(assemBytes)
    End Sub

#End Region

#Region "AssemblyResolveEventHandler Method"
    Function AssemblyResolveEventHandler(ByVal sender As Object, ByVal args As ResolveEventArgs) As [Assembly]
        ' Attempt to load assembly from m_projectOutputPath if it can't be found otherwise
        ' First, parse out assembly name from args.Name
        Dim tmpSplit() As String = args.Name.Split(","c)
        Dim asmName As String = tmpSplit(0)
        Dim asmBytes() As Byte = Nothing

        Dim loadedAsm As [Assembly] = Nothing
        Try
            asmBytes = LoadFile(m_projectOutputPath & asmName & ".dll")
        Catch
            ' try loading by .exe
            Try
                asmBytes = LoadFile(m_projectOutputPath & asmName & ".exe")
            Catch
                ' failed - return null
            End Try
        End Try
        If Not asmBytes Is Nothing Then
            loadedAsm = [Assembly].Load(asmBytes)
        End If
        Return loadedAsm
    End Function 'AssemblyResolveEventHandler

#End Region

    Private Function PerformCodePartLookup(ByVal psName As String, _
                                       ByVal poProjectImports As ArrayList, _
                                       ByVal poFileLevelImports As ArrayList, _
                                       ByVal poNamespace As Stack, _
                                       ByVal poClass As Stack, _
                                       ByVal poNamespaceCollection As NamespaceCollection, _
                                       ByVal pbOnlyTypes As Boolean) As CodePart


        Dim sTypeName As String
        Dim sNamespacePrefix As String
        Dim sClassPrefix As String
        Dim sNamePrefix As String
        Dim oCodePart As CodePart
        Dim iDotPos As Integer
        Dim sCurNamePrefix As String

        sNamespacePrefix = Utils.StringStackToSeparatedString(poNamespace)
        sClassPrefix = Utils.StringStackToSeparatedString(poClass)

        If (sNamespacePrefix.Length > 0) And (sClassPrefix.Length > 0) Then
            sNamePrefix = sNamespacePrefix + "." + sClassPrefix
        Else
            sNamePrefix = sNamespacePrefix + sClassPrefix
        End If

        ' File level imports:
        'Imports WindowsApplication2.NS1.NS2.NS3.class4
        'Imports Class5 = WindowsApplication2.NS1.NS2.NS3.class4

        ' 0) Forsk med fult navn
        oCodePart = GetCodePartFromNamespaceCollection(poNamespaceCollection, sTypeName, pbOnlyTypes)

        sCurNamePrefix = sNamePrefix

        If oCodePart Is Nothing Then
            

            While oCodePart Is Nothing And sCurNamePrefix.Length > 0
                oCodePart = GetCodePartFromNamespaceCollection(poNamespaceCollection, sCurNamePrefix + "." + psName, pbOnlyTypes)

                If oCodePart Is Nothing Then
                    iDotPos = sCurNamePrefix.LastIndexOf("."c)
                    If iDotPos <> -1 Then
                        sCurNamePrefix = sCurNamePrefix.Substring(0, iDotPos - 1)
                    Else
                        Exit While
                    End If
                End If
            End While
        End If

        If oCodePart Is Nothing Then
            ' 2) Forsk sEEbenytt aliaser
            oCodePart = GetCodePartFromNamespaceCollectionUsingImports(poNamespaceCollection, psName, poFileLevelImports, True, pbOnlyTypes)
        End If

        If oCodePart Is Nothing Then
            ' 3) Forsk sEEbenytt fil imports
            oCodePart = GetCodePartFromNamespaceCollectionUsingImports(poNamespaceCollection, psName, poFileLevelImports, False, pbOnlyTypes)
        End If

        If oCodePart Is Nothing Then
            ' 4) Forsk sEEbenytt prosjekt imports
            oCodePart = GetCodePartFromNamespaceCollectionUsingImports(poNamespaceCollection, psName, poProjectImports, False, pbOnlyTypes)
        End If

        Return oCodePart

    End Function

    Private Function GetCodePartFromNamespaceCollectionUsingImports(ByVal poNamespaceCollection As NamespaceCollection, ByVal psName As String, ByVal poImports As ArrayList, ByVal pbAliasMode As Boolean, ByVal pbOnlyTypes As Boolean) As CodePart

        Dim oImport As CodeImportStatement
        Dim oCodePart As CodePart

        If Not poImports Is Nothing Then
            For Each oImport In poImports
                If oImport.IsAlias = pbAliasMode Then
                    oCodePart = GetCodePartFromNamespaceCollection(poNamespaceCollection, oImport.Import + "." + psName, pbOnlyTypes)
                    If Not oCodePart Is Nothing Then
                        Exit For
                    End If
                End If
            Next
        End If

        Return oCodePart

    End Function

    Private Function GetCodePartFromNamespaceCollection(ByVal poNamespaceCollection As NamespaceCollection, ByVal psName As String, ByVal pbOnlyTypes As Boolean) As CodePart
        Dim oRetCodePart As CodePart
        Dim iLastPos As Integer
        Dim oType As Type

        Dim sNamespacePart As String
        Dim sTypePart As String
        Dim sMemberPart As String

        If oRetCodePart Is Nothing Then
            If Not pbOnlyTypes Then
                If poNamespaceCollection.Exists(psName) Then
                    oRetCodePart = New CodePart(psName)
                End If
            End If
        End If

        If oRetCodePart Is Nothing Then
            iLastPos = psName.LastIndexOf("."c)
            If iLastPos <> -1 Then
                ' TODO This does not work for .ctor does it???
                sNamespacePart = psName.Substring(0, iLastPos - 1)
                sTypePart = psName.Substring(iLastPos + 1)

                oType = GetTypeFromNamespace(poNamespaceCollection, sNamespacePart, sTypePart)
                If oType Is Nothing Then

                    If Not pbOnlyTypes Then
                        sMemberPart = sTypePart

                        iLastPos = sNamespacePart.LastIndexOf("."c)
                        If iLastPos <> -1 Then
                            sNamespacePart = psName.Substring(0, iLastPos - 1)
                            sTypePart = psName.Substring(iLastPos + 1)

                            oRetCodePart = GetTypeMemberOrMethodFromNamespace(poNamespaceCollection, sNamespacePart, sTypePart, sMemberPart)

                        End If
                    End If

                Else
                    oRetCodePart = New CodePart(oType)
                End If
            End If
        End If

        Return oRetCodePart

    End Function

    Private Function GetTypeFromNamespace(ByVal poNamespaceCollection As NamespaceCollection, ByVal psNamespace As String, ByVal psType As String) As Type

        Dim aAssemblies() As [Assembly]
        Dim oAssembly As [Assembly]
        Dim oRetType As Type

        aAssemblies = poNamespaceCollection.GetAssembliesForNamespace(psNamespace)

        If Not aAssemblies Is Nothing Then
            For Each oAssembly In aAssemblies
                oRetType = oAssembly.GetType(psNamespace + "." + psType, False, True)
                If Not oRetType Is Nothing Then
                    Exit For
                End If
            Next
        End If

        Return oRetType

    End Function

    Private Function GetTypeMemberOrMethodFromNamespace(ByVal poNamespaceCollection As NamespaceCollection, ByVal psNamespace As String, ByVal psType As String, ByVal psMemberOrMethod As String) As CodePart

        Dim oType As Type
        Dim oRetCodePart As CodePart

        Dim oMethod As MethodInfo
        Dim oProperty As Reflection.PropertyInfo
        Dim oEvent As Reflection.EventInfo
        Dim oField As Reflection.FieldInfo

        oType = GetTypeFromNamespace(poNamespaceCollection, psNamespace, psType)
        If Not oType Is Nothing Then

            If oRetCodePart Is Nothing Then
                For Each oMethod In oType.GetMethods(BindingFlags.Instance Or BindingFlags.NonPublic Or BindingFlags.Public Or BindingFlags.Static Or BindingFlags.FlattenHierarchy)
                    If oMethod.Name.ToLower = psMemberOrMethod.ToLower() Then
                        oRetCodePart = New CodePart(oMethod)
                    End If
                Next
            End If

            If oRetCodePart Is Nothing Then
                For Each oProperty In oType.GetProperties(BindingFlags.Instance Or BindingFlags.NonPublic Or BindingFlags.Public Or BindingFlags.Static Or BindingFlags.FlattenHierarchy)
                    If oProperty.Name.ToLower = psMemberOrMethod.ToLower() Then
                        oRetCodePart = New CodePart(oProperty)
                    End If
                Next
            End If

            If oRetCodePart Is Nothing Then
                For Each oEvent In oType.GetEvents(BindingFlags.Instance Or BindingFlags.NonPublic Or BindingFlags.Public Or BindingFlags.Static Or BindingFlags.FlattenHierarchy)
                    If oEvent.Name.ToLower = psMemberOrMethod.ToLower() Then
                        oRetCodePart = New CodePart(oEvent)
                    End If
                Next
            End If

            If oRetCodePart Is Nothing Then
                For Each oField In oType.GetFields(BindingFlags.Instance Or BindingFlags.NonPublic Or BindingFlags.Public Or BindingFlags.Static Or BindingFlags.FlattenHierarchy)
                    If oField.Name.ToLower = psMemberOrMethod.ToLower() Then
                        oRetCodePart = New CodePart(oField)
                    End If
                Next
            End If

            ' TODO what about .ctors?

        End If

        Return oRetCodePart

    End Function

End Class



